# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


import pyomo.common.unittest as unittest

from pyomo.environ import Var, Set, ConcreteModel, TransformationFactory
from pyomo.dae import ContinuousSet, DerivativeVar
from pyomo.dae.diffvar import DAE_Error

from io import StringIO

from pyomo.common.log import LoggingIntercept

from os.path import abspath, dirname, normpath, join

currdir = dirname(abspath(__file__))
exdir = normpath(join(currdir, '..', '..', '..', 'examples', 'dae'))


class TestFiniteDiff(unittest.TestCase):
    """
    Class for testing the pyomo.DAE finite difference discretization
    """

    def setUp(self):
        """
        Setting up testing model
        """
        self.m = m = ConcreteModel()
        m.t = ContinuousSet(bounds=(0, 10))
        m.v1 = Var(m.t)
        m.dv1 = DerivativeVar(m.v1)
        m.s = Set(initialize=[1, 2, 3], ordered=True)

    # test backward finite difference discretization
    # on var indexed by single ContinuousSet
    def test_disc_single_index_backward(self):
        m = self.m.clone()
        disc = TransformationFactory('dae.finite_difference')
        disc.apply_to(m, nfe=5)

        self.assertTrue(hasattr(m, 'dv1_disc_eq'))
        self.assertEqual(len(m.dv1_disc_eq), 5)
        self.assertEqual(len(m.v1), 6)

        expected_disc_points = [0, 2.0, 4.0, 6.0, 8.0, 10]
        disc_info = m.t.get_discretization_info()

        self.assertEqual(disc_info['scheme'], 'BACKWARD Difference')

        for idx, val in enumerate(list(m.t)):
            self.assertAlmostEqual(val, expected_disc_points[idx])

        self.assertTrue(hasattr(m, '_pyomo_dae_reclassified_derivativevars'))
        self.assertIn(m.dv1, m._pyomo_dae_reclassified_derivativevars)

        output = """\
dv1_disc_eq : Size=5, Index=t, Active=True
    Key : Lower : Body                               : Upper : Active
    2.0 :   0.0 :   dv1[2.0] - 0.5*(v1[2.0] - v1[0]) :   0.0 :   True
    4.0 :   0.0 : dv1[4.0] - 0.5*(v1[4.0] - v1[2.0]) :   0.0 :   True
    6.0 :   0.0 : dv1[6.0] - 0.5*(v1[6.0] - v1[4.0]) :   0.0 :   True
    8.0 :   0.0 : dv1[8.0] - 0.5*(v1[8.0] - v1[6.0]) :   0.0 :   True
     10 :   0.0 :   dv1[10] - 0.5*(v1[10] - v1[8.0]) :   0.0 :   True
"""
        out = StringIO()
        m.dv1_disc_eq.pprint(ostream=out)
        self.assertEqual(output, out.getvalue())

    # test backward finite difference discretization on
    # second order derivative indexed by single ContinuousSet
    def test_disc_second_order_backward(self):
        m = self.m.clone()
        m.dv1dt2 = DerivativeVar(m.v1, wrt=(m.t, m.t))
        disc = TransformationFactory('dae.finite_difference')
        disc.apply_to(m, nfe=2)

        self.assertTrue(hasattr(m, 'dv1dt2_disc_eq'))
        self.assertEqual(len(m.dv1dt2_disc_eq), 1)
        self.assertEqual(len(m.v1), 3)

        self.assertTrue(hasattr(m, '_pyomo_dae_reclassified_derivativevars'))
        self.assertIn(m.dv1, m._pyomo_dae_reclassified_derivativevars)
        self.assertIn(m.dv1dt2, m._pyomo_dae_reclassified_derivativevars)

        output = """\
dv1dt2_disc_eq : Size=1, Index=t, Active=True
    Key : Lower : Body                                           : Upper : Active
     10 :   0.0 : dv1dt2[10] - 0.04*(v1[10] - 2*v1[5.0] + v1[0]) :   0.0 :   True
"""
        out = StringIO()
        m.dv1dt2_disc_eq.pprint(ostream=out)
        self.assertEqual(output, out.getvalue())

    # test forward finite difference discretization
    # on var indexed by single ContinuousSet
    def test_disc_single_index_forward(self):
        m = self.m.clone()
        disc = TransformationFactory('dae.finite_difference')
        disc.apply_to(m, nfe=5, scheme='FORWARD')

        self.assertTrue(hasattr(m, 'dv1_disc_eq'))
        self.assertEqual(len(m.dv1_disc_eq), 5)
        self.assertEqual(len(m.v1), 6)

        expected_disc_points = [0, 2.0, 4.0, 6.0, 8.0, 10]
        disc_info = m.t.get_discretization_info()

        self.assertEqual(disc_info['scheme'], 'FORWARD Difference')

        for idx, val in enumerate(list(m.t)):
            self.assertAlmostEqual(val, expected_disc_points[idx])

        self.assertTrue(hasattr(m, '_pyomo_dae_reclassified_derivativevars'))
        self.assertIn(m.dv1, m._pyomo_dae_reclassified_derivativevars)

        output = """\
dv1_disc_eq : Size=5, Index=t, Active=True
    Key : Lower : Body                               : Upper : Active
      0 :   0.0 :     dv1[0] - 0.5*(v1[2.0] - v1[0]) :   0.0 :   True
    2.0 :   0.0 : dv1[2.0] - 0.5*(v1[4.0] - v1[2.0]) :   0.0 :   True
    4.0 :   0.0 : dv1[4.0] - 0.5*(v1[6.0] - v1[4.0]) :   0.0 :   True
    6.0 :   0.0 : dv1[6.0] - 0.5*(v1[8.0] - v1[6.0]) :   0.0 :   True
    8.0 :   0.0 :  dv1[8.0] - 0.5*(v1[10] - v1[8.0]) :   0.0 :   True
"""
        out = StringIO()
        m.dv1_disc_eq.pprint(ostream=out)
        self.assertEqual(output, out.getvalue())

    # test forward finite difference discretization
    # second order derivative indexed by single ContinuousSet
    def test_disc_second_order_forward(self):
        m = self.m.clone()
        m.dv1dt2 = DerivativeVar(m.v1, wrt=(m.t, m.t))
        disc = TransformationFactory('dae.finite_difference')
        disc.apply_to(m, nfe=2, scheme='FORWARD')

        self.assertTrue(hasattr(m, 'dv1dt2_disc_eq'))
        self.assertEqual(len(m.dv1dt2_disc_eq), 1)
        self.assertEqual(len(m.v1), 3)

        self.assertTrue(hasattr(m, '_pyomo_dae_reclassified_derivativevars'))
        self.assertIn(m.dv1, m._pyomo_dae_reclassified_derivativevars)
        self.assertIn(m.dv1dt2, m._pyomo_dae_reclassified_derivativevars)

        output = """\
dv1dt2_disc_eq : Size=1, Index=t, Active=True
    Key : Lower : Body                                          : Upper : Active
      0 :   0.0 : dv1dt2[0] - 0.04*(v1[10] - 2*v1[5.0] + v1[0]) :   0.0 :   True
"""
        out = StringIO()
        m.dv1dt2_disc_eq.pprint(ostream=out)
        self.assertEqual(output, out.getvalue())

    # test central finite difference discretization
    # on var indexed by single ContinuousSet
    def test_disc_single_index_central(self):
        m = self.m.clone()
        disc = TransformationFactory('dae.finite_difference')
        disc.apply_to(m, nfe=5, scheme='CENTRAL')

        self.assertTrue(hasattr(m, 'dv1_disc_eq'))
        self.assertEqual(len(m.dv1_disc_eq), 4)
        self.assertEqual(len(m.v1), 6)

        expected_disc_points = [0, 2.0, 4.0, 6.0, 8.0, 10]
        disc_info = m.t.get_discretization_info()

        self.assertEqual(disc_info['scheme'], 'CENTRAL Difference')

        for idx, val in enumerate(list(m.t)):
            self.assertAlmostEqual(val, expected_disc_points[idx])

        output = """\
dv1_disc_eq : Size=4, Index=t, Active=True
    Key : Lower : Body                                : Upper : Active
    2.0 :   0.0 :   dv1[2.0] - 0.25*(v1[4.0] - v1[0]) :   0.0 :   True
    4.0 :   0.0 : dv1[4.0] - 0.25*(v1[6.0] - v1[2.0]) :   0.0 :   True
    6.0 :   0.0 : dv1[6.0] - 0.25*(v1[8.0] - v1[4.0]) :   0.0 :   True
    8.0 :   0.0 :  dv1[8.0] - 0.25*(v1[10] - v1[6.0]) :   0.0 :   True
"""
        out = StringIO()
        m.dv1_disc_eq.pprint(ostream=out)
        self.assertEqual(output, out.getvalue())

    # test central finite difference discretization
    # second order derivative indexed by single ContinuousSet
    def test_disc_second_order_central(self):
        m = self.m.clone()
        m.dv1dt2 = DerivativeVar(m.v1, wrt=(m.t, m.t))
        disc = TransformationFactory('dae.finite_difference')
        disc.apply_to(m, nfe=2, scheme='CENTRAL')

        self.assertTrue(hasattr(m, 'dv1dt2_disc_eq'))
        self.assertEqual(len(m.dv1dt2_disc_eq), 1)
        self.assertEqual(len(m.v1), 3)

        output = """\
dv1dt2_disc_eq : Size=1, Index=t, Active=True
    Key : Lower : Body                                            : Upper : Active
    5.0 :   0.0 : dv1dt2[5.0] - 0.04*(v1[10] - 2*v1[5.0] + v1[0]) :   0.0 :   True
"""
        out = StringIO()
        m.dv1dt2_disc_eq.pprint(ostream=out)
        self.assertEqual(output, out.getvalue())

    # test collocation discretization on var indexed by ContinuousSet and Set
    def test_disc_multi_index(self):
        m = self.m.clone()
        m.v2 = Var(m.t, m.s)
        m.dv2 = DerivativeVar(m.v2)

        disc = TransformationFactory('dae.finite_difference')
        disc.apply_to(m, nfe=5)

        self.assertTrue(hasattr(m, 'dv1_disc_eq'))
        self.assertTrue(hasattr(m, 'dv2_disc_eq'))
        self.assertEqual(len(m.dv2_disc_eq), 15)
        self.assertEqual(len(m.v2), 18)

        expected_disc_points = [0, 2.0, 4.0, 6.0, 8.0, 10]
        disc_info = m.t.get_discretization_info()

        self.assertEqual(disc_info['scheme'], 'BACKWARD Difference')

        for idx, val in enumerate(list(m.t)):
            self.assertAlmostEqual(val, expected_disc_points[idx])

    # test collocation discretization on var indexed by multiple ContinuousSets
    def test_disc_multi_index2(self):
        m = self.m.clone()
        m.t2 = ContinuousSet(bounds=(0, 5))
        m.v2 = Var(m.t, m.t2)
        m.dv2dt = DerivativeVar(m.v2, wrt=m.t)
        m.dv2dt2 = DerivativeVar(m.v2, wrt=m.t2)

        disc = TransformationFactory('dae.finite_difference')
        disc.apply_to(m, nfe=2)

        self.assertTrue(hasattr(m, 'dv2dt_disc_eq'))
        self.assertTrue(hasattr(m, 'dv2dt2_disc_eq'))
        self.assertEqual(len(m.dv2dt_disc_eq), 6)
        self.assertEqual(len(m.dv2dt2_disc_eq), 6)
        self.assertEqual(len(m.v2), 9)

        expected_t_disc_points = [0, 5.0, 10]
        expected_t2_disc_points = [0, 2.5, 5]

        for idx, val in enumerate(list(m.t)):
            self.assertAlmostEqual(val, expected_t_disc_points[idx])

        for idx, val in enumerate(list(m.t2)):
            self.assertAlmostEqual(val, expected_t2_disc_points[idx])

    # test collocation discretization on var indexed by ContinuousSet and
    # multi-dimensional Set
    def test_disc_multidimen_index(self):
        m = self.m.clone()
        m.s2 = Set(initialize=[('A', 'B'), ('C', 'D'), ('E', 'F')])
        m.v2 = Var(m.t, m.s2)
        m.dv2 = DerivativeVar(m.v2)
        m.v3 = Var(m.s2, m.t)
        m.dv3 = DerivativeVar(m.v3)

        disc = TransformationFactory('dae.finite_difference')
        disc.apply_to(m, nfe=5)

        self.assertTrue(hasattr(m, 'dv1_disc_eq'))
        self.assertTrue(hasattr(m, 'dv2_disc_eq'))
        self.assertTrue(hasattr(m, 'dv3_disc_eq'))
        self.assertEqual(len(m.dv2_disc_eq), 15)
        self.assertEqual(len(m.v2), 18)
        self.assertEqual(len(m.dv3_disc_eq), 15)
        self.assertEqual(len(m.v3), 18)

        expected_disc_points = [0, 2.0, 4.0, 6.0, 8.0, 10]
        disc_info = m.t.get_discretization_info()

        self.assertEqual(disc_info['scheme'], 'BACKWARD Difference')

        for idx, val in enumerate(list(m.t)):
            self.assertAlmostEqual(val, expected_disc_points[idx])

    # test passing the discretization invalid options
    def test_disc_invalid_options(self):
        m = self.m.clone()

        with self.assertRaises(TypeError):
            TransformationFactory('dae.finite_difference').apply_to(m, wrt=m.s)

        with self.assertRaises(ValueError):
            TransformationFactory('dae.finite_difference').apply_to(m, nfe=-1)

        with self.assertRaises(ValueError):
            TransformationFactory('dae.finite_difference').apply_to(m, scheme='foo')

        with self.assertRaises(ValueError):
            TransformationFactory('dae.finite_difference').apply_to(m, foo=True)

        TransformationFactory('dae.finite_difference').apply_to(m, wrt=m.t)
        with self.assertRaises(ValueError):
            TransformationFactory('dae.finite_difference').apply_to(m, wrt=m.t)

        m = self.m.clone()
        disc = TransformationFactory('dae.finite_difference')
        disc.apply_to(m)
        with self.assertRaises(ValueError):
            disc.apply_to(m)

    # test discretization using fewer points than ContinuousSet initialized
    # with
    def test_initialized_continuous_set(self):
        m = ConcreteModel()
        m.t = ContinuousSet(initialize=[0, 1, 2, 3, 4])
        m.v = Var(m.t)
        m.dv = DerivativeVar(m.v)

        log_out = StringIO()
        with LoggingIntercept(log_out, 'pyomo.dae'):
            TransformationFactory('dae.finite_difference').apply_to(m, nfe=2)
        self.assertIn('More finite elements', log_out.getvalue())

    # Test discretizing an invalid derivative
    def test_invalid_derivative(self):
        m = ConcreteModel()
        m.t = ContinuousSet(bounds=(0, 10))
        m.v = Var(m.t)
        m.dv = DerivativeVar(m.v, wrt=(m.t, m.t, m.t))

        with self.assertRaises(DAE_Error):
            TransformationFactory('dae.finite_difference').apply_to(m)

    # test trying to discretize a ContinuousSet twice
    def test_discretize_twice(self):
        m = self.m.clone()

        disc1 = TransformationFactory('dae.finite_difference')
        disc1.apply_to(m, nfe=5)

        disc2 = TransformationFactory('dae.finite_difference')

        with self.assertRaises(DAE_Error):
            disc2.apply_to(m, nfe=5)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.piecewise.piecewise_linear_expression import (
    PiecewiseLinearExpression,
)
from pyomo.core import Expression
from pyomo.core.expr.visitor import StreamBasedExpressionVisitor


class PiecewiseLinearToMIP(StreamBasedExpressionVisitor):
    """
    Expression walker to replace PiecewiseLinearExpressions when creating
    equivalent MIP formulations.

    Args:
        transform_pw_linear_expression (function): a callback that accepts
            a PiecewiseLinearExpression, its parent PiecewiseLinearFunction,
            and a transformation Block. It is expected to convert the
            PiecewiseLinearExpression to MIP form, and return the Var (or
            other expression) that will replace it in the expression.
        transBlock (Block): transformation Block to pass to the above
            callback
    """

    def __init__(self, transform_pw_linear_expression, transBlock):
        self.transform_pw_linear_expression = transform_pw_linear_expression
        self.transBlock = transBlock
        self._process_node = self._process_node_bx

    def initializeWalker(self, expr):
        expr, src, src_idx = expr
        # always walk
        return True, expr

    def beforeChild(self, node, child, child_idx):
        return True, None

    def exitNode(self, node, data):
        if node.__class__ is PiecewiseLinearExpression:
            parent = node.pw_linear_function
            substitute_var = self.transform_pw_linear_expression(
                node, parent, self.transBlock
            )
            parent._expressions[parent._expression_ids[node]] = substitute_var
        return node

    finalizeResult = None



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


from .common.factory import SolverFactory
from .solvers.ipopt import Ipopt, LegacyIpoptSolver
from .solvers.gurobi.gurobi_direct import GurobiDirect
from .solvers.gurobi.gurobi_persistent import GurobiPersistent
from .solvers.gurobi.gurobi_direct_minlp import GurobiDirectMINLP
from .solvers.highs import Highs
from .solvers.gams import GAMS
from .solvers.knitro.direct import KnitroDirectSolver


def load():
    SolverFactory.register(
        name="ipopt", legacy_name="ipopt_v2", doc="The IPOPT NLP solver"
    )(Ipopt, LegacyIpoptSolver)
    SolverFactory.register(
        name="gurobi_persistent",
        legacy_name="gurobi_persistent_v2",
        doc="Persistent interface to Gurobi",
    )(GurobiPersistent)
    SolverFactory.register(
        name="gurobi_direct",
        legacy_name="gurobi_direct_v2",
        doc="Direct (scipy-based) interface to Gurobi",
    )(GurobiDirect)
    SolverFactory.register(
        name='gurobi_direct_minlp',
        legacy_name='gurobi_direct_minlp',
        doc='Direct interface to Gurobi accommodating general MINLP',
    )(GurobiDirectMINLP)
    SolverFactory.register(
        name="highs", legacy_name="highs", doc="Persistent interface to HiGHS"
    )(Highs)
    SolverFactory.register(name='gams', legacy_name='gams_v2', doc='Interface to GAMS')(
        GAMS
    )
    SolverFactory.register(
        name="knitro_direct",
        legacy_name="knitro_direct",
        doc="Direct interface to KNITRO solver",
    )(KnitroDirectSolver)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________


import datetime
import time
import io
from operator import attrgetter, itemgetter

from pyomo.common.dependencies import attempt_import
from pyomo.common.collections import ComponentMap, ComponentSet
from pyomo.common.config import ConfigDict, ConfigValue
from pyomo.common.errors import InvalidValueError
from pyomo.common.numeric_types import native_complex_types
from pyomo.common.shutdown import python_is_shutting_down
from pyomo.common.timing import HierarchicalTimer

from pyomo.contrib.solver.common.factory import SolverFactory
from pyomo.contrib.solver.common.solution_loader import SolutionLoaderBase
from pyomo.contrib.solver.common.util import NoSolutionError
from .gurobi_direct_base import GurobiDirectBase, GurobiDirectSolutionLoaderBase

from pyomo.core.base import (
    Binary,
    Block,
    BooleanVar,
    Constraint,
    Expression,
    Integers,
    minimize,
    maximize,
    NonNegativeIntegers,
    NonNegativeReals,
    NonPositiveIntegers,
    NonPositiveReals,
    Objective,
    Param,
    Reals,
    SortComponents,
    Suffix,
    Var,
    value,
)
import pyomo.core.expr as EXPR
from pyomo.core.expr.numeric_expr import (
    NegationExpression,
    ProductExpression,
    DivisionExpression,
    PowExpression,
    AbsExpression,
    UnaryFunctionExpression,
    Expr_ifExpression,
    LinearExpression,
    MonomialTermExpression,
    SumExpression,
)
from pyomo.core.expr.visitor import StreamBasedExpressionVisitor, _EvaluationVisitor
from pyomo.core.staleflag import StaleFlagManager

from pyomo.opt import WriterFactory
from pyomo.repn.quadratic import QuadraticRepnVisitor
from pyomo.repn.util import (
    apply_node_operation,
    categorize_valid_components,
    ExprType,
    ExitNodeDispatcher,
    BeforeChildDispatcher,
    complex_number_error,
    initialize_exit_node_dispatcher,
    nan,
    OrderedVarRecorder,
    check_constant,
)

import sys

### FIXME: Remove the following as soon as non-active components no
### longer report active==True
from pyomo.network import Port
from pyomo.core.base import RangeSet, Set

###

_CONSTANT = ExprType.CONSTANT
_GENERAL = ExprType.GENERAL
_LINEAR = ExprType.LINEAR
_QUADRATIC = ExprType.QUADRATIC
_VARIABLE = ExprType.VARIABLE

_function_map = {}


def _finalize_gurobipy(gurobipy, available):
    if not available:
        return
    _function_map.update(
        {
            'exp': (_GENERAL, gurobipy.nlfunc.exp),
            'log': (_GENERAL, gurobipy.nlfunc.log),
            'log10': (_GENERAL, gurobipy.nlfunc.log10),
            'sin': (_GENERAL, gurobipy.nlfunc.sin),
            'cos': (_GENERAL, gurobipy.nlfunc.cos),
            'tan': (_GENERAL, gurobipy.nlfunc.tan),
            'sqrt': (_GENERAL, gurobipy.nlfunc.sqrt),
            # Not supporting any of these right now--we'd have to build them from the
            # above:
            # 'asin': None,
            # 'sinh': None,
            # 'asinh': None,
            # 'acos': None,
            # 'cosh': None,
            # 'acosh': None,
            # 'atan': None,
            # 'tanh': None,
            # 'atanh': None,
            # 'ceil': None,
            # 'floor': None,
        }
    )


gurobipy, gurobipy_available = attempt_import(
    'gurobipy',
    deferred_submodules=['GRB'],
    callback=_finalize_gurobipy,
    minimum_version='12.0.0',
)
GRB = gurobipy.GRB


"""
In Gurobi 12:

If you have f(x) == 0, you must write it as z == f(x) and then write z == 0.
Basically, you must introduce auxiliary variables for all the general nonlinear
parts. (And no worries about additively separable or anything--they do that 
under the hood).

In this implementation, we replace the *entire* LHS of the constraint with the
auxiliary variable rather than just the nonlinear part. Otherwise we would really
need to keep track of what nonlinear subexpressions we had already replaced and make
sure to use the same auxiliary variables, and from what we know, this is probably not
worth it.

We are not using Gurobi's '.nl' attribute at all for now--its usage seems like the
exception rather than the rule, so we will let Gurobi expand the expressions for now.
"""


def _create_grb_var(visitor, pyomo_var, name=""):
    pyo_domain = pyomo_var.domain
    domain_lb, domain_ub, domain = pyo_domain.get_interval()
    if domain not in (0, 1):
        raise ValueError(
            "Unsupported domain for Var '%s': %s" % (pyomo_var.name, pyomo_var.domain)
        )
    domain = GRB.INTEGER if domain else GRB.CONTINUOUS
    # We set binaries to be binary right now because we don't know if Gurbi cares
    if pyo_domain is Binary:
        domain = GRB.BINARY

    # returns tigter of bounds from domain and bounds set on variable
    lb, ub = pyomo_var.bounds
    if lb is None:
        lb = -float("inf")
    if ub is None:
        ub = float("inf")

    return visitor.grb_model.addVar(lb=lb, ub=ub, vtype=domain, name=name)


class GurobiMINLPBeforeChildDispatcher(BeforeChildDispatcher):
    @staticmethod
    def _before_var(visitor, child):
        if child not in visitor.var_map:
            if child.fixed:
                return False, (_CONSTANT, check_constant(child.value, child, visitor))
            grb_var = _create_grb_var(
                visitor,
                child,
                name=child.name if visitor.symbolic_solver_labels else "",
            )
            visitor.var_map[child] = grb_var
        return False, (_VARIABLE, visitor.var_map[child])

    @staticmethod
    def _before_named_expression(visitor, child):
        _id = id(child)
        if _id in visitor.subexpression_cache:
            _type, expr = visitor.subexpression_cache[_id]
            return False, (_type, expr)
        else:
            return True, None


def _handle_node_with_eval_expr_visitor_invariant(visitor, node, data):
    """
    Calls expression evaluation visitor on nodes that have an invariant
    expression type in the return.
    """
    return (data[0], visitor._eval_expr_visitor.visit(node, (data[1],)))


def _handle_node_with_eval_expr_visitor_unknown(visitor, node, *data):
    # The expression type is whatever the highest one of the incoming arguments
    # was.
    expr_type = max(map(itemgetter(0), data))
    return (
        expr_type,
        visitor._eval_expr_visitor.visit(node, tuple(map(itemgetter(1), data))),
    )


def _handle_node_with_eval_expr_visitor_constant(visitor, node, *data):
    return (
        _CONSTANT,
        visitor._eval_expr_visitor.visit(node, tuple(map(itemgetter(1), data))),
    )


def _handle_node_with_eval_expr_visitor_linear(visitor, node, *data):
    return (
        _LINEAR,
        visitor._eval_expr_visitor.visit(node, tuple(map(itemgetter(1), data))),
    )


def _handle_node_with_eval_expr_visitor_quadratic(visitor, node, *data):
    return (
        _QUADRATIC,
        visitor._eval_expr_visitor.visit(node, tuple(map(itemgetter(1), data))),
    )


def _handle_node_with_eval_expr_visitor_nonlinear(visitor, node, *data):
    # ESJ: _apply_operation for DivisionExpression expects that result
    # supports __getitem__, so I'm expanding the map to a tuple.
    return (
        _GENERAL,
        visitor._eval_expr_visitor.visit(node, tuple(map(itemgetter(1), data))),
    )


def _handle_linear_constant_pow_expr(visitor, node, arg1, arg2):
    expr_type = _GENERAL
    if arg2[1] == 1:
        expr_type = _LINEAR
    if arg2[1] == 2:
        expr_type = _QUADRATIC
    return (
        expr_type,
        visitor._eval_expr_visitor.visit(node, tuple(map(itemgetter(1), (arg1, arg2)))),
    )


def _handle_quadratic_constant_pow_expr(visitor, node, arg1, arg2):
    expr_type = _GENERAL
    if arg2[1] == 1:
        expr_type = _QUADRATIC
    return (
        expr_type,
        visitor._eval_expr_visitor.visit(node, tuple(map(itemgetter(1), (arg1, arg2)))),
    )


def _handle_unary(visitor, node, data):
    if node._name in _function_map:
        expr_type, fcn = _function_map[node._name]
        return expr_type, fcn(data[1])
    raise ValueError(
        "The unary function '%s' is not supported by the Gurobi MINLP writer."
        % node._name
    )


def _handle_unary_constant(visitor, node, data):
    try:
        return _CONSTANT, node._fcn(value(data[1]))
    except:
        raise InvalidValueError(
            f"Invalid number encountered evaluating constant unary expression "
            f"{node}: {sys.exc_info()[1]}"
        )


def _handle_named_expression(visitor, node, arg1):
    # Record this common expression
    visitor.subexpression_cache[id(node)] = arg1
    _type, arg1 = arg1
    return _type, arg1


def _handle_abs_constant(visitor, node, arg1):
    return (_CONSTANT, abs(arg1[1]))


def _handle_abs_var(visitor, node, arg1):
    # This auxiliary variable actually is non-negative, yay absolute value!
    aux_abs = visitor.grb_model.addVar()
    visitor.grb_model.addConstr(aux_abs == gurobipy.abs_(arg1[1]))

    return (_VARIABLE, aux_abs)


def _handle_abs_expression(visitor, node, arg1):
    # we need an auxiliary variable
    aux_arg = visitor.grb_model.addVar(lb=-GRB.INFINITY, ub=GRB.INFINITY)
    visitor.grb_model.addConstr(aux_arg == arg1[1])
    # This one truly is non-negative because it's an absolute value
    aux_abs = visitor.grb_model.addVar()
    visitor.grb_model.addConstr(aux_abs == gurobipy.abs_(aux_arg))

    return (_VARIABLE, aux_abs)


def define_exit_node_handlers(_exit_node_handlers=None):
    if _exit_node_handlers is None:
        _exit_node_handlers = {}

    # We can rely on operator overloading for many, but not all expressions.
    _exit_node_handlers[SumExpression] = {
        None: _handle_node_with_eval_expr_visitor_unknown
    }
    _exit_node_handlers[LinearExpression] = {
        # Can come back LINEAR or CONSTANT, so we use the 'unknown' version
        None: _handle_node_with_eval_expr_visitor_unknown
    }
    _exit_node_handlers[NegationExpression] = {
        None: _handle_node_with_eval_expr_visitor_invariant
    }
    _exit_node_handlers[ProductExpression] = {
        None: _handle_node_with_eval_expr_visitor_nonlinear,
        (_CONSTANT, _CONSTANT): _handle_node_with_eval_expr_visitor_constant,
        (_CONSTANT, _LINEAR): _handle_node_with_eval_expr_visitor_linear,
        (_CONSTANT, _QUADRATIC): _handle_node_with_eval_expr_visitor_quadratic,
        (_CONSTANT, _VARIABLE): _handle_node_with_eval_expr_visitor_linear,
        (_LINEAR, _CONSTANT): _handle_node_with_eval_expr_visitor_linear,
        (_LINEAR, _LINEAR): _handle_node_with_eval_expr_visitor_quadratic,
        (_LINEAR, _VARIABLE): _handle_node_with_eval_expr_visitor_quadratic,
        (_VARIABLE, _CONSTANT): _handle_node_with_eval_expr_visitor_linear,
        (_VARIABLE, _LINEAR): _handle_node_with_eval_expr_visitor_quadratic,
        (_VARIABLE, _VARIABLE): _handle_node_with_eval_expr_visitor_quadratic,
    }
    _exit_node_handlers[MonomialTermExpression] = _exit_node_handlers[ProductExpression]
    _exit_node_handlers[DivisionExpression] = {
        None: _handle_node_with_eval_expr_visitor_nonlinear,
        (_CONSTANT, _CONSTANT): _handle_node_with_eval_expr_visitor_constant,
        (_LINEAR, _CONSTANT): _handle_node_with_eval_expr_visitor_linear,
        (_VARIABLE, _CONSTANT): _handle_node_with_eval_expr_visitor_linear,
        (_QUADRATIC, _CONSTANT): _handle_node_with_eval_expr_visitor_quadratic,
    }
    _exit_node_handlers[PowExpression] = {
        None: _handle_node_with_eval_expr_visitor_nonlinear,
        (_CONSTANT, _CONSTANT): _handle_node_with_eval_expr_visitor_constant,
        (_VARIABLE, _CONSTANT): _handle_linear_constant_pow_expr,
        (_LINEAR, _CONSTANT): _handle_linear_constant_pow_expr,
        (_QUADRATIC, _CONSTANT): _handle_quadratic_constant_pow_expr,
    }
    _exit_node_handlers[UnaryFunctionExpression] = {
        None: _handle_unary,
        (_CONSTANT,): _handle_unary_constant,
    }

    ## TODO: ExprIf, RangedExpressions (if we do exprif...)
    _exit_node_handlers[Expression] = {None: _handle_named_expression}

    # These are special because of quirks of Gurobi's current support for general
    # nonlinear:
    _exit_node_handlers[AbsExpression] = {
        None: _handle_abs_expression,
        (_CONSTANT,): _handle_abs_constant,
        (_VARIABLE,): _handle_abs_var,
    }

    return _exit_node_handlers


class GurobiMINLPVisitor(StreamBasedExpressionVisitor):
    before_child_dispatcher = GurobiMINLPBeforeChildDispatcher()
    exit_node_dispatcher = ExitNodeDispatcher(
        initialize_exit_node_dispatcher(define_exit_node_handlers())
    )

    def __init__(self, grb_model, symbolic_solver_labels=False):
        super().__init__()
        self.grb_model = grb_model
        self.symbolic_solver_labels = symbolic_solver_labels
        self.var_map = ComponentMap()
        self.subexpression_cache = {}
        self._eval_expr_visitor = _EvaluationVisitor(True)
        self.evaluate = self._eval_expr_visitor.dfs_postorder_stack

    def initializeWalker(self, expr):
        walk, result = self.beforeChild(None, expr, 0)
        if not walk:
            return False, self.finalizeResult(result)
        return True, expr

    def beforeChild(self, node, child, child_idx):
        return self.before_child_dispatcher[child.__class__](self, child)

    def exitNode(self, node, data):
        return self.exit_node_dispatcher[(node.__class__, *map(itemgetter(0), data))](
            self, node, *data
        )

    def finalizeResult(self, result):
        return result


@WriterFactory.register(
    'gurobi_minlp',
    'Direct interface to Gurobi that allows for general nonlinear expressions',
)
class GurobiMINLPWriter:
    CONFIG = ConfigDict('gurobi_minlp_writer')
    CONFIG.declare(
        'symbolic_solver_labels',
        ConfigValue(
            default=False,
            domain=bool,
            description='Write Pyomo Var and Constraint names to Gurobi model',
        ),
    )

    def __init__(self):
        self.config = self.CONFIG()

    def _create_gurobi_expression(
        self, expr, src, src_index, grb_model, quadratic_visitor, grb_visitor
    ):
        """
        Returns a gurobipy representation of the expression
        """
        expr_type, grb_expr = grb_visitor.walk_expression(expr)
        if expr_type is not _GENERAL:
            return expr_type, grb_expr, False, None
        else:
            aux = grb_model.addVar(lb=-GRB.INFINITY, ub=GRB.INFINITY)
            return expr_type, grb_expr, True, aux

    def write(self, model, **options):
        config = options.pop('config', self.config)(options)

        components, unknown = categorize_valid_components(
            model,
            active=True,
            sort=SortComponents.deterministic,
            valid={
                Block,
                Expression,
                Var,
                BooleanVar,
                Param,
                Suffix,
                # FIXME: Non-active components should not report as Active
                Set,
                RangeSet,
                Port,
            },
            targets={Objective, Constraint},
        )
        if unknown:
            raise ValueError(
                "The model ('%s') contains the following active components "
                "that the Gurobi MINLP writer does not know how to "
                "process:\n\t%s"
                % (
                    model.name,
                    "\n\t".join(
                        sorted(
                            "%s:\n\t\t%s"
                            % (k, "\n\t\t".join(sorted(map(attrgetter('name'), v))))
                            for k, v in unknown.items()
                        )
                    ),
                )
            )

        # Get a quadratic walker instance
        quadratic_visitor = QuadraticRepnVisitor(
            subexpression_cache={}, var_recorder=OrderedVarRecorder({}, {}, None)
        )

        # create Gurobi model
        grb_model = gurobipy.Model()
        visitor = GurobiMINLPVisitor(
            grb_model, symbolic_solver_labels=config.symbolic_solver_labels
        )

        active_objs = []
        if components[Objective]:
            for block in components[Objective]:
                for obj in block.component_data_objects(
                    Objective,
                    active=True,
                    descend_into=False,
                    sort=SortComponents.deterministic,
                ):
                    active_objs.append(obj)
        if len(active_objs) > 1:
            raise ValueError(
                "More than one active objective defined for "
                "input model '%s': Cannot write to gurobipy." % model.name
            )
        elif len(active_objs) == 1:
            obj = active_objs[0]
            pyo_obj = [obj]
            if obj.sense is minimize:
                sense = GRB.MINIMIZE
            else:
                sense = GRB.MAXIMIZE
            expr_type, obj_expr, nonlinear, aux = self._create_gurobi_expression(
                obj.expr, obj, 0, grb_model, quadratic_visitor, visitor
            )
            if nonlinear:
                # The objective must be linear or quadratic, so we move the nonlinear
                # one to the constraints
                grb_model.setObjective(aux, sense=sense)
                grb_model.addConstr(aux == obj_expr)
            else:
                grb_model.setObjective(obj_expr, sense=sense)
        # else it's fine--Gurobi doesn't require us to give an objective, so we don't
        # either, but we do have to pass the info through for the results object
        else:
            pyo_obj = []

        # write constraints
        pyo_cons = []
        grb_cons = []

        if components[Constraint]:
            for block in components[Constraint]:
                for cons in block.component_data_objects(
                    Constraint,
                    active=True,
                    descend_into=False,
                    sort=SortComponents.deterministic,
                ):
                    lb, body, ub = cons.to_bounded_expression(evaluate_bounds=True)
                    expr_type, expr, nonlinear, aux = self._create_gurobi_expression(
                        body, cons, 0, grb_model, quadratic_visitor, visitor
                    )
                    if nonlinear:
                        grb_model.addConstr(aux == expr)
                        expr = aux
                    elif expr_type == _CONSTANT:
                        # cast everything to a float in case there are numpy
                        # types because you can't do addConstr(np.True_)
                        expr = float(expr)
                        if lb is not None:
                            lb = float(lb)
                        if ub is not None:
                            ub = float(ub)
                    if cons.equality:
                        grb_cons.append(grb_model.addConstr(expr == lb))
                        pyo_cons.append(cons)
                    else:
                        # TODO: should we have special handling if expr is a
                        # GRB.LinExpr so that we can use the ranged linear
                        # constraint syntax (expr == [lb, ub])?
                        if lb is not None:
                            grb_cons.append(grb_model.addConstr(expr >= lb))
                            pyo_cons.append(cons)
                        if ub is not None:
                            grb_cons.append(grb_model.addConstr(expr <= ub))
                            pyo_cons.append(cons)

        grb_model.update()
        return grb_model, visitor.var_map, pyo_obj, grb_cons, pyo_cons


class GurobiDirectMINLPSolutionLoader(GurobiDirectSolutionLoaderBase):
    def __init__(self, solver_model, var_map, con_map) -> None:
        super().__init__(solver_model)
        self._var_map = var_map
        self._con_map = con_map

    def _get_var_lists(self):
        return list(self._var_map.keys()), list(self._var_map.values())

    def _get_var_map(self):
        return self._var_map

    def _get_con_map(self):
        return self._con_map

    def __del__(self):
        super().__del__()
        if python_is_shutting_down():
            return
        # Free the associated model
        if self._solver_model is not None:
            self._var_map = None
            self._con_map = None
            # explicitly release the model
            self._solver_model.dispose()
            self._solver_model = None


@SolverFactory.register(
    'gurobi_direct_minlp',
    doc='Direct interface to Gurobi version 12 and up '
    'supporting general nonlinear expressions',
)
class GurobiDirectMINLP(GurobiDirectBase):
    _minimum_version = (12, 0, 0)

    def __init__(self, **kwds):
        super().__init__(**kwds)
        self._var_map = None

    def _pyomo_gurobi_var_iter(self):
        return self._var_map.items()

    def _create_solver_model(self, pyomo_model, config):
        timer = config.timer
        timer.start('compile_model')

        writer = GurobiMINLPWriter()
        grb_model, var_map, pyo_obj, grb_cons, pyo_cons = writer.write(
            pyomo_model, symbolic_solver_labels=config.symbolic_solver_labels
        )

        timer.stop('compile_model')

        self._var_map = var_map
        con_map = {}
        for pc, gc in zip(pyo_cons, grb_cons):
            if pc in con_map:
                # range constraint
                con_map[pc] = (con_map[pc], gc)
            else:
                con_map[pc] = gc

        solution_loader = GurobiDirectMINLPSolutionLoader(
            solver_model=grb_model, var_map=var_map, con_map=con_map
        )

        return grb_model, solution_loader, bool(pyo_obj)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import re
import sys
import os
import subprocess

import pyomo.common
from pyomo.common.errors import ApplicationError

import pyomo.opt.solver.shellcmd
from pyomo.opt.solver.shellcmd import SystemCallSolver


class ILMLicensedSystemCallSolver(SystemCallSolver):
    """A command line solver that launches executables licensed with ILM"""

    def __init__(self, **kwds):
        """Constructor"""
        pyomo.opt.solver.shellcmd.SystemCallSolver.__init__(self, **kwds)

    def available(self, exception_flag=False):
        """True if the solver is available"""
        if self._assert_available:
            return True
        if not pyomo.opt.solver.shellcmd.SystemCallSolver.available(
            self, exception_flag
        ):
            return False
        executable = pyomo.common.Executable("ilmlist")
        if executable:
            try:
                cmd = [executable.path()]
                if sys.platform[0:3] == "win":
                    # on windows, the ilm license manager by default
                    # pauses after displaying the token status, so that
                    # the window doesn't disappear and the user can
                    # actually read it. however, if we don't suppress
                    # this behavior, this command will stall until the
                    # user hits Ctrl-C.
                    cmd.append("-batch")
                result = subprocess.run(
                    cmd,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    universal_newlines=True,
                )
            except OSError:
                msg = sys.exc_info()[1]
                raise ApplicationError(
                    "Could not execute the command: ilmtest\n\tError message: " + msg
                )
            sys.stdout.flush()
            for line in result.stdout.split("\n"):
                tokens = re.split('[\t ]+', line.strip())
                if (
                    len(tokens) == 5
                    and tokens[0] == 'tokens'
                    and tokens[1] == 'reserved:'
                    and tokens[4] == os.environ.get('USER', None)
                ):
                    if not (tokens[2] == 'none' or tokens[2] == '0'):
                        return True
                    return False
                elif (
                    len(tokens) == 3
                    and tokens[0] == 'available'
                    and tokens[1] == 'tokens:'
                ):
                    if tokens[2] == '0':
                        return False
                    break
                elif (
                    len(tokens) == 6 and tokens[1] == 'server' and tokens[5] == 'DOWN.'
                ):
                    return False
        return True



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Development of this module was conducted as part of the Institute for
# the Design of Advanced Energy Systems (IDAES) with support through the
# Simulation-Based Engineering, Crosscutting Research Program within the
# U.S. Department of Energy's Office of Fossil Energy and Carbon Management.
# ____________________________________________________________________________________

import logging

from pyomo.core.base.range import NumericRange
from pyomo.common.config import (
    ConfigDict,
    ConfigValue,
    Bool,
    PositiveInt,
    PositiveFloat,
    In,
    document_kwargs_from_configdict,
)
from pyomo.contrib.trustregion.filter import Filter, FilterElement
from pyomo.contrib.trustregion.interface import TRFInterface
from pyomo.contrib.trustregion.util import IterationLogger
from pyomo.opt import SolverFactory

logger = logging.getLogger('pyomo.contrib.trustregion')

__version__ = (0, 2, 0)


def trust_region_method(model, decision_variables, ext_fcn_surrogate_map_rule, config):
    """
    The main driver of the Trust Region algorithm method.

    Parameters
    ----------
    model : ConcreteModel
        The user's model to be solved.
    degrees_of_freedom_variables : List of Vars
        User-supplied input. The user must provide a list of vars which
        are the degrees of freedom or decision variables within
        the model.
    ext_fcn_surrogate_map_rule : Function, optional
        In the 2020 Yoshio/Biegler paper, this is referred to as
        the basis function `b(w)`.
        This is the low-fidelity model with which to solve the original
        process model problem and which is integrated into the
        surrogate model.
        The default is 0 (i.e., no basis function rule.)
    config : ConfigDict
        This holds the solver and TRF-specific configuration options.

    """

    # Initialize necessary TRF methods
    TRFLogger = IterationLogger()
    TRFilter = Filter()
    interface = TRFInterface(
        model, decision_variables, ext_fcn_surrogate_map_rule, config
    )

    # Initialize the problem
    rebuildSM = False
    obj_val, feasibility = interface.initializeProblem()
    # Initialize first iteration feasibility/objective value to enable
    # termination check
    feasibility_k = feasibility
    obj_val_k = obj_val
    # Initialize step_norm_k to a bogus value to enable termination check
    step_norm_k = 0
    # Initialize trust region radius
    trust_radius = config.trust_radius

    iteration = 0

    TRFLogger.newIteration(
        iteration, feasibility_k, obj_val_k, trust_radius, step_norm_k
    )
    TRFLogger.logIteration()
    if config.verbose:
        TRFLogger.printIteration()
    while iteration < config.maximum_iterations:
        iteration += 1

        # Check termination conditions
        if (feasibility_k <= config.feasibility_termination) and (
            step_norm_k <= config.step_size_termination
        ):
            print('EXIT: Optimal solution found.')
            interface.model.display()
            break

        # If trust region very small and no progress is being made,
        # terminate. The following condition must hold for two
        # consecutive iterations.
        if (trust_radius <= config.minimum_radius) and (
            abs(feasibility_k - feasibility) < config.feasibility_termination
        ):
            if subopt_flag:
                logger.warning('WARNING: Insufficient progress.')
                print('EXIT: Feasible solution found.')
                break
            else:
                subopt_flag = True
        else:
            # This condition holds for iteration 0, which will declare
            # the boolean subopt_flag
            subopt_flag = False

        # Set bounds to enforce the trust region
        interface.updateDecisionVariableBounds(trust_radius)
        # Generate suggorate model r_k(w)
        if rebuildSM:
            interface.updateSurrogateModel()

        # Solve the Trust Region Subproblem (TRSP)
        obj_val_k, step_norm_k, feasibility_k = interface.solveModel()

        TRFLogger.newIteration(
            iteration, feasibility_k, obj_val_k, trust_radius, step_norm_k
        )

        # Check filter acceptance
        filterElement = FilterElement(obj_val_k, feasibility_k)
        if not TRFilter.isAcceptable(filterElement, config.maximum_feasibility):
            # Reject the step
            TRFLogger.iterrecord.rejected = True
            trust_radius = max(
                config.minimum_radius, step_norm_k * config.radius_update_param_gamma_c
            )
            rebuildSM = False
            interface.rejectStep()
            # Log iteration information
            TRFLogger.logIteration()
            if config.verbose:
                TRFLogger.printIteration()
            continue

        # Switching condition: Eq. (7) in Yoshio/Biegler (2020)
        if (obj_val - obj_val_k) >= (
            config.switch_condition_kappa_theta
            * pow(feasibility, config.switch_condition_gamma_s)
        ) and (feasibility <= config.minimum_feasibility):
            # f-type step
            TRFLogger.iterrecord.fStep = True
            trust_radius = min(
                max(step_norm_k * config.radius_update_param_gamma_e, trust_radius),
                config.maximum_radius,
            )
        else:
            # theta-type step
            TRFLogger.iterrecord.thetaStep = True
            filterElement = FilterElement(
                obj_val_k - config.param_filter_gamma_f * feasibility_k,
                (1 - config.param_filter_gamma_theta) * feasibility_k,
            )
            TRFilter.addToFilter(filterElement)
            # Calculate ratio: Eq. (10) in Yoshio/Biegler (2020)
            rho_k = (
                feasibility - feasibility_k + config.feasibility_termination
            ) / max(feasibility, config.feasibility_termination)
            # Ratio tests: Eq. (8) in Yoshio/Biegler (2020)
            # If rho_k is between eta_1 and eta_2, trust radius stays same
            if (rho_k < config.ratio_test_param_eta_1) or (
                feasibility > config.minimum_feasibility
            ):
                trust_radius = max(
                    config.minimum_radius,
                    (config.radius_update_param_gamma_c * step_norm_k),
                )
            elif rho_k >= config.ratio_test_param_eta_2:
                trust_radius = min(
                    config.maximum_radius,
                    max(
                        trust_radius, (config.radius_update_param_gamma_e * step_norm_k)
                    ),
                )

        TRFLogger.updateIteration(trustRadius=trust_radius)
        # Accept step and reset for next iteration
        rebuildSM = True
        feasibility = feasibility_k
        obj_val = obj_val_k
        # Log iteration information
        TRFLogger.logIteration()
        if config.verbose:
            TRFLogger.printIteration()

    if iteration >= config.maximum_iterations:
        logger.warning(
            'EXIT: Maximum iterations reached: {}.'.format(config.maximum_iterations)
        )

    return interface.model


def _trf_config():
    """
    Generate the configuration dictionary.
    The user may change the configuration options during the instantiation
    of the trustregion solver:
        >>> optTRF = SolverFactory('trustregion',
        ...                        solver='ipopt',
        ...                        maximum_iterations=50,
        ...                        minimum_radius=1e-5,
        ...                        verbose=True)

    The user may also update the configuration after instantiation:
        >>> optTRF = SolverFactory('trustregion')
        >>> optTRF._CONFIG.trust_radius = 0.5

    The user may also update the configuration as part of the solve call:
        >>> optTRF = SolverFactory('trustregion')
        >>> optTRF.solve(model, decision_variables, trust_radius=0.5)
    Returns
    -------
    CONFIG : ConfigDict
        This holds all configuration options to be passed to the TRF solver.

    """
    CONFIG = ConfigDict('TrustRegion')

    ### Solver options
    CONFIG.declare(
        'solver',
        ConfigValue(default='ipopt', description='Solver to use. Default = ``ipopt``.'),
    )
    CONFIG.declare(
        'keepfiles',
        ConfigValue(
            default=False,
            domain=Bool,
            description="Optional. Whether or not to "
            "write files of sub-problems for use in debugging. "
            "Default = False.",
        ),
    )
    CONFIG.declare(
        'tee',
        ConfigValue(
            default=False,
            domain=Bool,
            description="Optional. Sets the ``tee`` "
            "for sub-solver(s) utilized. "
            "Default = False.",
        ),
    )

    ### Trust Region specific options
    CONFIG.declare(
        'verbose',
        ConfigValue(
            default=False,
            domain=Bool,
            description="Optional. When True, print each "
            "iteration's relevant information to the console "
            "as well as to the log. "
            "Default = False.",
        ),
    )
    CONFIG.declare(
        'trust_radius',
        ConfigValue(
            default=1.0,
            domain=PositiveFloat,
            description="Initial trust region radius ``delta_0``. Default = 1.0.",
        ),
    )
    CONFIG.declare(
        'minimum_radius',
        ConfigValue(
            default=1e-6,
            domain=PositiveFloat,
            description="Minimum allowed trust region radius ``delta_min``. "
            "Default = 1e-6.",
        ),
    )
    CONFIG.declare(
        'maximum_radius',
        ConfigValue(
            default=CONFIG.trust_radius * 100,
            domain=PositiveFloat,
            description="Maximum allowed trust region radius. If trust region "
            "radius reaches maximum allowed, solver will exit. "
            "Default = 100 * trust_radius.",
        ),
    )
    CONFIG.declare(
        'maximum_iterations',
        ConfigValue(
            default=50,
            domain=PositiveInt,
            description="Maximum allowed number of iterations. Default = 50.",
        ),
    )
    ### Termination options
    CONFIG.declare(
        'feasibility_termination',
        ConfigValue(
            default=1e-5,
            domain=PositiveFloat,
            description="Feasibility measure termination tolerance ``epsilon_theta``. "
            "Default = 1e-5.",
        ),
    )
    CONFIG.declare(
        'step_size_termination',
        ConfigValue(
            default=CONFIG.feasibility_termination,
            domain=PositiveFloat,
            description="Step size termination tolerance ``epsilon_s``. "
            "Matches the feasibility termination tolerance by default.",
        ),
    )
    ### Switching Condition options
    CONFIG.declare(
        'minimum_feasibility',
        ConfigValue(
            default=1e-4,
            domain=PositiveFloat,
            description="Minimum feasibility measure ``theta_min``. Default = 1e-4.",
        ),
    )
    CONFIG.declare(
        'switch_condition_kappa_theta',
        ConfigValue(
            default=0.1,
            domain=In(NumericRange(0, 1, 0, (False, False))),
            description="Switching condition parameter ``kappa_theta``. "
            "Contained in open set (0, 1). "
            "Default = 0.1.",
        ),
    )
    CONFIG.declare(
        'switch_condition_gamma_s',
        ConfigValue(
            default=2.0,
            domain=PositiveFloat,
            description="Switching condition parameter ``gamma_s``. "
            "Must satisfy: ``gamma_s > 1/(1+mu)`` where ``mu`` "
            "is contained in set (0, 1]. "
            "Default = 2.0.",
        ),
    )
    ### Trust region update/ratio test parameters
    CONFIG.declare(
        'radius_update_param_gamma_c',
        ConfigValue(
            default=0.5,
            domain=In(NumericRange(0, 1, 0, (False, False))),
            description="Lower trust region update parameter ``gamma_c``. "
            "Default = 0.5.",
        ),
    )
    CONFIG.declare(
        'radius_update_param_gamma_e',
        ConfigValue(
            default=2.5,
            domain=In(NumericRange(1, None, 0)),
            description="Upper trust region update parameter ``gamma_e``. "
            "Default = 2.5.",
        ),
    )
    CONFIG.declare(
        'ratio_test_param_eta_1',
        ConfigValue(
            default=0.05,
            domain=In(NumericRange(0, 1, 0, (False, False))),
            description="Lower ratio test parameter ``eta_1``. "
            "Must satisfy: ``0 < eta_1 <= eta_2 < 1``. "
            "Default = 0.05.",
        ),
    )
    CONFIG.declare(
        'ratio_test_param_eta_2',
        ConfigValue(
            default=0.2,
            domain=In(NumericRange(0, 1, 0, (False, False))),
            description="Lower ratio test parameter ``eta_2``. "
            "Must satisfy: ``0 < eta_1 <= eta_2 < 1``. "
            "Default = 0.2.",
        ),
    )
    ### Filter
    CONFIG.declare(
        'maximum_feasibility',
        ConfigValue(
            default=50.0,
            domain=PositiveFloat,
            description="Maximum allowable feasibility measure ``theta_max``. "
            "Parameter for use in filter method."
            "Default = 50.0.",
        ),
    )
    CONFIG.declare(
        'param_filter_gamma_theta',
        ConfigValue(
            default=0.01,
            domain=In(NumericRange(0, 1, 0, (False, False))),
            description="Fixed filter parameter ``gamma_theta`` within (0, 1). "
            "Default = 0.01",
        ),
    )
    CONFIG.declare(
        'param_filter_gamma_f',
        ConfigValue(
            default=0.01,
            domain=In(NumericRange(0, 1, 0, (False, False))),
            description="Fixed filter parameter ``gamma_f`` within (0, 1). "
            "Default = 0.01",
        ),
    )

    return CONFIG


@SolverFactory.register(
    'trustregion',
    doc='Trust region algorithm "solver" for black box/glass box optimization',
)
class TrustRegionSolver:
    """
    The Trust Region Solver is a 'solver' based on the 2016/2018/2020 AiChE
    papers by Eason (2016/2018), Yoshio (2020), and Biegler.

    """

    CONFIG = _trf_config()

    def __init__(self, **kwds):
        self.config = self.CONFIG(kwds)

    def available(self, exception_flag=True):
        """
        Check if solver is available.
        """
        return True

    def version(self):
        """
        Return a 3-tuple describing the solver version.
        """
        return __version__

    def license_is_valid(self):
        """
        License for using Trust Region solver.
        """
        return True

    def __enter__(self):
        return self

    def __exit__(self, et, ev, tb):
        pass

    @document_kwargs_from_configdict(CONFIG)
    def solve(
        self,
        model,
        degrees_of_freedom_variables,
        ext_fcn_surrogate_map_rule=None,
        **kwds,
    ):
        """
        This method calls the TRF algorithm.

        Parameters
        ----------
        model : ConcreteModel
            The model to be solved using the Trust Region Framework.
        degrees_of_freedom_variables : List[Var]
            User-supplied input. The user must provide a list of vars which
            are the degrees of freedom or decision variables within
            the model.
        ext_fcn_surrogate_map_rule : Function, optional
            In the 2020 Yoshio/Biegler paper, this is referred to as
            the basis function `b(w)`.
            This is the low-fidelity model with which to solve the original
            process model problem and which is integrated into the
            surrogate model.
            The default is 0 (i.e., no basis function rule.)

        """
        config = self.config(kwds.pop('options', {}))
        config.set_value(kwds)
        if ext_fcn_surrogate_map_rule is None:
            # If the user does not pass us a "basis" function,
            # we default to 0.
            ext_fcn_surrogate_map_rule = lambda comp, ef: 0
        result = trust_region_method(
            model, degrees_of_freedom_variables, ext_fcn_surrogate_map_rule, config
        )
        return result



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

# """
# Useful functions for transformations.
# """

from inspect import isroutine
from pyomo.core import Var, Objective, Constraint, Set, Param


def collectAbstractComponents(model):
    """
    Returns all abstract constraints, objectives, variables,
    parameters, and sets. Does not query instance values, if
    present. Returns nested dictionaries describing the model,
    including all rules, domains, bounds, and index sets.
    """

    # Clone the model for sanity
    cp = model.clone()

    #
    # Component Name
    #   - Field in map
    #
    # Constraint
    #   - Index set
    #   - Rule
    #
    # Var
    #   - Index set
    #   - Domain
    #   - Bounds
    #
    # Objective
    #   - Index set
    #   - Rule
    #
    # Set
    #   - Index set
    #   - Initialize
    #   - Domain
    #
    # Param
    #   - Index set
    #   - initialize
    #   - domain/bounds
    #
    # All components are primarily indexed by their attribute name (string)
    #

    # Model components (will include all subclasses of the base class)
    constraints = {}
    variables = {}
    objectives = {}
    parameters = {}
    sets = {}

    # Mapping naming scheme
    conName = "Constraint"
    varName = "Var"
    objName = "Objective"
    paramName = "Param"
    setName = "Set"
    index = "index"
    bounds = "bounds"
    domain = "domain"
    initialize = "initialize"
    rule = "rule"

    # Iterate over all model components
    for comp in cp._ctypes:
        # Collect all Constraint objects
        if issubclass(comp, Constraint):
            comps = cp.component_map(comp, active=True)
            for name, obj in [(name, comps[name]) for name in comps]:
                # Query this constraint's attributes
                data = {}

                # Get indices
                data[index] = _getAbstractIndices(obj)

                # Get the rule
                data[rule] = _getAbstractRule(obj)

                # Add this constraint
                constraints[name] = data

        # Collect all Objective objects
        if issubclass(comp, Objective):
            comps = cp.component_map(comp, active=True)
            for name, obj in [(name, comps[name]) for name in comps]:
                # Query this objective's attributes
                data = {}

                # Get indices
                data[index] = _getAbstractIndices(obj)

                # Get the rule
                data[rule] = _getAbstractRule(obj)

                # Add this constraint
                objectives[name] = data

        # Collect all Var objects
        if issubclass(comp, Var):
            comps = cp.component_map(comp, active=True)
            for name, obj in [(name, comps[name]) for name in comps]:
                # Query this variable's attributes
                data = {}

                # Get indices
                data[index] = _getAbstractIndices(obj)

                # Get the domain
                data[domain] = _getAbstractDomain(obj)

                # Get the bounds
                data[bounds] = _getAbstractBounds(obj)

                # Add this constraint
                variables[name] = data

        # Collect all Set objects
        if issubclass(comp, Set):
            comps = cp.component_map(comp, active=True)
            for name, obj in [(name, comps[name]) for name in comps]:
                # Query this variable's attributes
                data = {}

                # Get indices
                data[index] = _getAbstractIndices(obj)

                # Get the domain
                data[domain] = _getAbstractDomain(obj)

                # Add this constraint
                sets[name] = data

        # Collect all Param objects
        if issubclass(comp, Param):
            comps = cp.component_map(comp, active=True)
            for name, obj in [(name, comps[name]) for name in comps]:
                # Query this variable's attributes
                data = {}

                # Get indices
                data[index] = _getAbstractIndices(obj)

                # Get the domain
                data[domain] = _getAbstractDomain(obj)

                # Add this constraint
                parameters[name] = data

    # Construct master data set
    master = {}
    master[conName] = constraints
    master[objName] = objectives
    master[varName] = variables
    master[paramName] = parameters
    master[setName] = sets

    return master


def _getAbstractIndices(comp):
    """
    Returns the index or index set of this component
    """
    if type(comp.index_set()) != type({}):
        # Singly indexed component
        return comp.index_set()
    else:
        # Unindexed constraint
        return {None: None}


def _getAbstractRule(comp):
    """
    Returns the rule defining this component
    """
    return comp.rule


def _getAbstractDomain(comp):
    """
    Returns the domain of this component
    """
    return getattr(comp, 'domain', None)


def _getAbstractBounds(comp):
    """
    Returns the bounds of this component
    """
    if getattr(comp, 'bounds', None) is None:
        return (None, None)
    else:
        return comp.bounds


def _getAbstractInitialize(comp):
    """
    Returns the initialization rule. If initialize is a container; return None;
    that information will be collected during construction.
    """
    if isroutine(comp.initialize):
        return comp.initialize
    else:
        return None


try:
    from functools import partial as _partial
except ImportError:
    # functools doesn't exist in Python 2.4
    def _partial(f, *args, **kwds):
        """
        Returns a new function with positional and keyword arguments
        partially applied

        """

        def closure(*cargs, **ckwds):
            # Collect positional arguments
            tmp_args = list(args)
            tmp_args.extend(list(cargs))

            tmp_kwds = dict(kwds)
            tmp_kwds.update(ckwds)

            # Call the original function
            return f(*tmp_args, **tmp_kwds)

        return closure


def partial(*args, **kwargs):
    """
    copy.deepcopy balks at copying anonymous functions. This overrides
    the default behavior of functools.partial to make deepcopy return
    the function itself, rather than attempting to copy it.
    """
    func = _partial(*args, **kwargs)

    def _partial_deepcopy(memo={}):
        return func

    func.__deepcopy__ = _partial_deepcopy
    return func


def process_canonical_repn(expr):
    """
    Returns a dictionary of {var_name_or_None: coef} values. None
    indicates a numeric constant.
    """

    terms = {}

    # Get the variables from the canonical representation
    vars = expr.pop(-1, {})

    # Find the linear terms
    linear = expr.pop(1, {})
    for k in linear:
        # FrozenDicts don't support (k, v)-style iteration
        v = linear[k]

        # There's exactly 1 variable in each term
        terms[vars[k.keys()[0]].label] = v

    # Get the constant term, if present
    const = expr.pop(0, {})
    if None in const:
        terms[None] = const[None]

    if len(expr) != 0:
        raise TypeError("Nonlinear terms in expression")

    return terms



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest

from pyomo.core import ConcreteModel, Var, Param, Constraint, Objective, exp
from pyomo.repn.standard_repn import generate_standard_repn as gar


class AmplRepnTests(unittest.TestCase):
    def test_divide_by_mutable(self):
        #
        # Test from https://github.com/Pyomo/pyomo/issues/153
        #
        m = ConcreteModel()
        m.x = Var(bounds=(1, 5))
        m.p = Param(initialize=100, mutable=True)
        m.con = Constraint(expr=exp(5 * (1 / m.x - 1 / m.p)) <= 10)
        m.obj = Objective(expr=m.x**2)

        test = gar(m.con.body)
        self.assertEqual(test.constant, 0)
        self.assertEqual(test.linear_vars, tuple())
        self.assertEqual(test.linear_coefs, tuple())
        self.assertEqual(set(id(v) for v in test.nonlinear_vars), set([id(m.x)]))
        self.assertIs(test.nonlinear_expr, m.con.body)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import sys
import pprint as _pprint_

from pyomo.common.collections import ComponentMap
import pyomo.core
from pyomo.core.expr.numvalue import NumericValue
from pyomo.core.kernel.base import (
    ICategorizedObject,
    _no_ctype,
    _convert_ctype,
    _convert_descend_into,
)


def preorder_traversal(node, ctype=_no_ctype, active=True, descend=True):
    """
    A generator that yields each object in the storage tree
    (including the root object) using a preorder traversal.

    Args:
        node: The root object.
        ctype: Indicates the category of components to
            include. The default value indicates that all
            categories should be included.
        active (:const:`True`/:const:`None`): Controls
            whether or not to filter the iteration to
            include only the active part of the storage
            tree. The default is :const:`True`. Setting this
            keyword to :const:`None` causes the active
            status of objects to be ignored.
        descend (bool, function): Controls if a container
            object should be descended into during the
            traversal. When a function is supplied, each
            container object will be passed into it and the
            return value will indicate if the traversal
            continues into children of the
            container. Default is True, which is equivalent
            to `lambda x: True`.

    Returns:
        iterator of objects in the storage tree, including
        the root object
    """
    assert active in (None, True)

    # if not active, then nothing below is active
    if (active is not None) and (not node.active):
        return

    # convert AML types into Kernel types (hack for the
    # solver interfaces)
    ctype = _convert_ctype.get(ctype, ctype)

    # convert descend to a function
    descend = _convert_descend_into(descend)

    if (
        (ctype is _no_ctype)
        or (node.ctype is ctype)
        or (node.ctype._is_heterogeneous_container)
    ):
        yield node

    if (not node._is_container) or (not descend(node)):
        return

    for child in node.children():
        child_ctype = child.ctype
        if not child._is_container:
            # not a container
            if (active is None) or child.active:
                if (ctype is _no_ctype) or (child_ctype is ctype):
                    yield child
        elif child._is_heterogeneous_container:
            # a heterogeneous container, so use
            # its traversal method
            for obj in preorder_traversal(
                child, ctype=ctype, active=active, descend=descend
            ):
                yield obj
        else:
            # a homogeneous container
            if child_ctype._is_heterogeneous_container:
                # this function ensures that the user provided
                # descend function is not called twice
                # on heterogeneous containers
                def descend_(obj_):
                    if obj_._is_heterogeneous_container:
                        return False
                    else:
                        return descend(obj_)

                for obj in preorder_traversal(child, active=active, descend=descend_):
                    if not obj._is_heterogeneous_container:
                        yield obj
                    else:
                        # a heterogeneous container, so use
                        # its traversal method and reapply the
                        # ctype filter
                        for item in preorder_traversal(
                            obj, ctype=ctype, active=active, descend=descend
                        ):
                            yield item
            elif (ctype is _no_ctype) or (child_ctype is ctype):
                for obj in preorder_traversal(child, active=active, descend=descend):
                    yield obj


def generate_names(node, convert=str, prefix="", **kwds):
    """
    Generate names relative to this object for all
    objects stored under it.

    This function is useful in situations where names
    are used often, but they do not need to be
    dynamically regenerated each time.

    Args:
        node: The root object below which names are
            generated.
        convert (function): A function that converts a
            storage key into a string
            representation. Default is str.
        prefix (str): A string to prefix names with.
        **kwds: Additional keywords passed to the
            preorder_traversal function.

    Returns:
        A component map that behaves as a dictionary
        mapping objects to names.
    """
    traversal = preorder_traversal(node, **kwds)
    names = ComponentMap()

    # skip the root object
    try:
        next(traversal)
    except StopIteration:
        # might be an empty traversal
        return names

    for obj in traversal:
        parent = obj.parent
        name = parent._child_storage_entry_string % convert(obj.storage_key)
        if parent is not node:
            names[obj] = names[parent] + parent._child_storage_delimiter_string + name
        else:
            names[obj] = prefix + name

    return names


def pprint(obj, indent=0, stream=sys.stdout):
    """pprint a kernel modeling object"""
    if not isinstance(obj, ICategorizedObject):
        if isinstance(obj, NumericValue):
            prefix = ""
            if indent > 0:
                prefix = (" " * indent) + " - "
            stream.write(prefix + str(obj) + "\n")
        else:
            assert indent == 0
            _pprint_.pprint(obj, indent=indent + 1, stream=stream)
        return
    if not obj._is_container:
        prefix = ""
        if indent > 0:
            prefix = (" " * indent) + " - "
        # not a block
        clsname = obj.__class__.__name__
        if obj.ctype is pyomo.core.kernel.variable.IVariable:
            stream.write(
                prefix
                + "%s: %s(active=%s, value=%s, bounds=(%s,%s), domain_type=%s, fixed=%s, stale=%s)\n"
                % (
                    str(obj),
                    clsname,
                    obj.active,
                    obj.value,
                    obj.lb,
                    obj.ub,
                    obj.domain_type.__name__,
                    obj.fixed,
                    obj.stale,
                )
            )
        elif obj.ctype is pyomo.core.kernel.constraint.IConstraint:
            stream.write(
                prefix
                + "%s: %s(active=%s, expr=%s)\n"
                % (str(obj), clsname, obj.active, str(obj.expr))
            )
        elif obj.ctype is pyomo.core.kernel.objective.IObjective:
            stream.write(
                prefix
                + "%s: %s(active=%s, expr=%s)\n"
                % (str(obj), clsname, obj.active, str(obj.expr))
            )
        elif obj.ctype is pyomo.core.kernel.expression.IExpression:
            stream.write(
                prefix
                + "%s: %s(active=%s, expr=%s)\n"
                % (str(obj), clsname, obj.active, str(obj.expr))
            )
        elif obj.ctype is pyomo.core.kernel.parameter.IParameter:
            stream.write(
                prefix
                + "%s: %s(active=%s, value=%s)\n"
                % (str(obj), clsname, obj.active, str(obj()))
            )
        elif obj.ctype is pyomo.core.kernel.sos.ISOS:
            stream.write(
                prefix
                + "%s: %s(active=%s, level=%s, entries=%s)\n"
                % (
                    str(obj),
                    clsname,
                    obj.active,
                    obj.level,
                    str(
                        [
                            "(%s,%s)" % (str(v), w)
                            for v, w in zip(obj.variables, obj.weights)
                        ]
                    ),
                )
            )
        else:
            assert obj.ctype is pyomo.core.kernel.suffix.ISuffix
            stream.write(
                prefix
                + "%s: %s(active=%s, size=%s)\n"
                % (str(obj.name), clsname, obj.active, str(len(obj)))
            )
    else:
        prefix = ""
        if indent > 0:
            prefix = (" " * indent) + " - "
        stream.write(
            prefix
            + "%s: %s(active=%s, ctype=%s)\n"
            % (str(obj), obj.__class__.__name__, obj.active, obj.ctype.__name__)
        )
        for c in obj.children():
            pprint(c, indent=indent + 1, stream=stream)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
from pyomo.common import unittest
import pyomo.environ as pyo
from pyomo.core.expr.numeric_expr import (
    LinearExpression,
    MonomialTermExpression,
    SumExpression,
    ProductExpression,
    DivisionExpression,
    PowExpression,
    NegationExpression,
    UnaryFunctionExpression,
    ExternalFunctionExpression,
    Expr_ifExpression,
    AbsExpression,
)
from pyomo.core.expr.relational_expr import (
    InequalityExpression,
    EqualityExpression,
    RangedExpression,
)
from pyomo.core.expr.compare import (
    convert_expression_to_prefix_notation,
    compare_expressions,
    assertExpressionsEqual,
    assertExpressionsStructurallyEqual,
)
from pyomo.common.gsl import find_GSL


class TestConvertToPrefixNotation(unittest.TestCase):
    def test_linear_expression(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var([1, 2, 3, 4])
        e = LinearExpression(
            constant=3, linear_coefs=list(m.x.keys()), linear_vars=list(m.x.values())
        )
        expected = [
            (LinearExpression, 5),
            3,
            (MonomialTermExpression, 2),
            1,
            m.x[1],
            (MonomialTermExpression, 2),
            2,
            m.x[2],
            (MonomialTermExpression, 2),
            3,
            m.x[3],
            (MonomialTermExpression, 2),
            4,
            m.x[4],
        ]
        pn = convert_expression_to_prefix_notation(e)
        self.assertEqual(pn, expected)

    def test_multiple(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()

        e = m.x**2 + m.x * m.y / 3 + 4
        expected = [
            (SumExpression, 3),
            (PowExpression, 2),
            m.x,
            2,
            (DivisionExpression, 2),
            (ProductExpression, 2),
            m.x,
            m.y,
            3,
            4,
        ]
        pn = convert_expression_to_prefix_notation(e)
        self.assertEqual(pn, expected)
        e2 = m.x**2 + m.x * m.y / 3 + 4
        e3 = m.y**2 + m.x * m.y / 3 + 4
        self.assertTrue(compare_expressions(e, e2))
        self.assertFalse(compare_expressions(e, e3))

    def test_unary(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        e = pyo.log(m.x)
        expected = [(UnaryFunctionExpression, 1, 'log'), m.x]
        pn = convert_expression_to_prefix_notation(e)
        self.assertEqual(expected, pn)

    def test_external_function(self):
        DLL = find_GSL()
        if not DLL:
            self.skipTest('Could not find the amplgsl.dll library')

        m = pyo.ConcreteModel()
        m.hypot = pyo.ExternalFunction(library=DLL, function='gsl_hypot')
        m.x = pyo.Var(initialize=0.5)
        m.y = pyo.Var(initialize=1.5)
        e = 2 * m.hypot(m.x, m.x * m.y)
        expected = [
            (ProductExpression, 2),
            2,
            (ExternalFunctionExpression, 2, m.hypot),
            m.x,
            (ProductExpression, 2),
            m.x,
            m.y,
        ]
        pn = convert_expression_to_prefix_notation(e)
        self.assertEqual(expected, pn)

    def test_var(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        pn = convert_expression_to_prefix_notation(m.x)
        self.assertEqual(pn, [m.x])

    def test_float(self):
        pn = convert_expression_to_prefix_notation(4.3)
        self.assertEqual(pn, [4.3])

    def test_monomial(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        e = 2 * m.x
        pn = convert_expression_to_prefix_notation(e)
        expected = [(MonomialTermExpression, 2), 2, m.x]
        self.assertEqual(pn, expected)

    def test_negation(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        e = -m.x**2
        pn = convert_expression_to_prefix_notation(e)
        expected = [(NegationExpression, 1), (PowExpression, 2), m.x, 2]
        self.assertEqual(pn, expected)

    def test_abs(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        e = abs(m.x)
        pn = convert_expression_to_prefix_notation(e)
        expected = [(AbsExpression, 1, 'abs'), m.x]
        self.assertEqual(pn, expected)

    def test_expr_if(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.y = pyo.Var()
        e = pyo.Expr_if(m.x <= 0, m.y + m.x == 0, m.y - m.x == 0)
        pn = convert_expression_to_prefix_notation(e)
        expected = [
            (Expr_ifExpression, 3),
            (InequalityExpression, 2, False),
            m.x,
            0,
            (EqualityExpression, 2),
            (LinearExpression, 2),
            m.y,
            m.x,
            0,
            (EqualityExpression, 2),
            (LinearExpression, 2),
            m.y,
            (MonomialTermExpression, 2),
            -1,
            m.x,
            0,
        ]
        self.assertEqual(pn, expected)

    def test_ranged_expression(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        e = pyo.inequality(-1, m.x, 1)
        pn = convert_expression_to_prefix_notation(e)
        expected = [(RangedExpression, 3, (False, False)), -1, m.x, 1]
        self.assertEqual(pn, expected)

    def test_assertExpressionsEqual(self):
        m = pyo.ConcreteModel()
        m.x = pyo.Var()
        m.e1 = pyo.Expression(expr=m.x**2 + m.x - 1)
        m.e2 = pyo.Expression(expr=m.x**2 + m.x - 1)
        m.f = pyo.Expression(expr=m.x**2 + 2 * m.x - 1)
        m.g = pyo.Expression(expr=m.x**2 + m.x - 2)

        assertExpressionsEqual(self, m.e1.expr, m.e2.expr)
        assertExpressionsStructurallyEqual(self, m.e1.expr, m.e2.expr)
        with self.assertRaisesRegex(AssertionError, 'Expressions not equal:'):
            assertExpressionsEqual(self, m.e1.expr, m.f.expr)
        with self.assertRaisesRegex(
            AssertionError, 'Expressions not structurally equal:'
        ):
            assertExpressionsStructurallyEqual(self, m.e1.expr, m.f.expr)

        # Structurally equal will compare across clones, whereas strict
        # equality will not
        i = m.clone()
        with self.assertRaisesRegex(AssertionError, 'Expressions not equal:'):
            assertExpressionsEqual(self, m.e1.expr, i.e1.expr)
        assertExpressionsStructurallyEqual(self, m.e1.expr, i.e1.expr)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core import ConcreteModel, Var, Objective, Piecewise
from pyomo.solvers.tests.models.base import _BaseTestModel, register_model


@register_model
class LP_piecewise(_BaseTestModel):
    """
    A continuous linear model
    """

    description = "LP_piecewise"
    capabilities = set(['linear'])

    def __init__(self):
        _BaseTestModel.__init__(self)
        self.add_results(self.description + ".json")

    def _generate_model(self):
        self.model = ConcreteModel()
        model = self.model
        model._name = self.description

        model.x = Var()
        model.y = Var()

        model.obj = Objective(expr=model.y)
        model.p = Piecewise(
            model.y,
            model.x,
            pw_pts=[-1, 0, 1],
            f_rule=[1, 0.5, 1],
            pw_repn='SOS2',
            pw_constr_type='LB',
            unbounded_domain_var=True,
        )

    def warmstart_model(self):
        assert self.model is not None
        model = self.model
        model.x.value = None
        model.y.value = 1.0


@register_model
class LP_piecewise_nosuffixes(LP_piecewise):
    description = "LP_piecewise_nosuffixes"
    test_pickling = False

    def __init__(self):
        LP_piecewise.__init__(self)
        self.disable_suffix_tests = True
        self.add_results("LP_piecewise.json")



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.core.base.constraint import Constraint
from pyomo.core.base.set import Set


def get_piecewise_constant_constraints(inputs, time, sample_points, use_next=True):
    """Returns an IndexedConstraint that constrains the provided variables
    to be constant between the provided sample points

    Arguments
    ---------
    inputs: list of variables
        Time-indexed variables that will be constrained piecewise constant
    time: Set
        Set of points at which provided variables will be constrained
    sample_points: List of floats
        Points at which "constant constraints" will be omitted; these are
        points at which the provided variables may vary.
    use_next: Bool (default True)
        Whether the next time point will be used in the constant constraint
        at each point in time. Otherwise, the previous time point is used.

    Returns
    -------
    Set, IndexedConstraint
        A RangeSet indexing the list of variables provided and a Constraint
        indexed by the product of this RangeSet and time.

    """
    input_set = Set(initialize=range(len(inputs)))
    sample_point_set = set(sample_points)

    def piecewise_constant_rule(m, i, t):
        if t in sample_point_set:
            return Constraint.Skip
        else:
            # I think whether we want prev or next here depends on whether
            # we use an explicit or implicit time discretization. I.e. whether
            # an input is applied to the finite element in front of or behind
            # its time point. If the wrong direction for a discretization
            # is used, we could have different inputs applied within the same
            # finite element, which I think we never want.
            var = inputs[i]
            if use_next:
                t_next = time.next(t)
                return var[t] - var[t_next] == 0
            else:
                t_prev = time.prev(t)
                return var[t_prev] - var[t] == 0

    pwc_con = Constraint(input_set, time, rule=piecewise_constant_rule)
    return input_set, pwc_con



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
This module redefines the global help() function.  If the 'thing'
being requested for help has the '__help__' attribute, then that
string is used to provide help information.  Otherwise, the standard
help utility is used.
"""

old_help = None
# there are times when the help function is not available
# (e.g., after 'freezing' code into an executable with tools
# like py2exe or cx_Freeze)
try:
    old_help = help
except NameError:
    old_help = None


def help(thing=None):  # pragma:nocover
    if not thing is None and hasattr(thing, '__help__'):
        print(thing.__help__)
    else:
        if old_help is None:
            raise NameError("Builtin 'help' is not available")
        old_help(thing)


try:
    __builtins__['help'] = help
except:  # pragma:nocover
    # If this fails, then just die silently.
    pass



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.appsi.base import (
    PersistentBase,
    PersistentSolver,
    SolverConfig,
    Results,
    TerminationCondition,
    PersistentSolutionLoader,
)
from pyomo.core.expr.numeric_expr import (
    ProductExpression,
    DivisionExpression,
    PowExpression,
    SumExpression,
    MonomialTermExpression,
    NegationExpression,
    UnaryFunctionExpression,
    LinearExpression,
    AbsExpression,
    NPV_ProductExpression,
    NPV_DivisionExpression,
    NPV_PowExpression,
    NPV_SumExpression,
    NPV_NegationExpression,
    NPV_UnaryFunctionExpression,
    NPV_AbsExpression,
)
from pyomo.common.errors import PyomoException
from pyomo.common.collections import ComponentMap
from pyomo.core.expr.numvalue import native_numeric_types
from typing import Dict, Optional, List
from pyomo.core.base.block import BlockData
from pyomo.core.base.var import VarData
from pyomo.core.base.param import ParamData
from pyomo.core.base.constraint import ConstraintData
from pyomo.common.timing import HierarchicalTimer
from pyomo.core.base import SymbolMap, NumericLabeler, TextLabeler
from pyomo.common.dependencies import attempt_import
from pyomo.core.staleflag import StaleFlagManager
from pyomo.contrib.appsi.cmodel import cmodel, cmodel_available

wntr, wntr_available = attempt_import('wntr')
import logging
import time
import sys
from pyomo.core.expr.visitor import ExpressionValueVisitor

logger = logging.getLogger(__name__)


class WntrConfig(SolverConfig):
    def __init__(
        self,
        description=None,
        doc=None,
        implicit=False,
        implicit_domain=None,
        visibility=0,
    ):
        super().__init__(
            description=description,
            doc=doc,
            implicit=implicit,
            implicit_domain=implicit_domain,
            visibility=visibility,
        )


class WntrResults(Results):
    def __init__(self, solver):
        super().__init__()
        self.wallclock_time = None
        self.solution_loader = PersistentSolutionLoader(solver=solver)


class Wntr(PersistentBase, PersistentSolver):
    def __init__(self, only_child_vars=True):
        super().__init__(only_child_vars=only_child_vars)
        self._config = WntrConfig()
        self._solver_options = dict()
        self._solver_model = None
        self._symbol_map = SymbolMap()
        self._labeler = None
        self._pyomo_var_to_solver_var_map = dict()
        self._pyomo_con_to_solver_con_map = dict()
        self._pyomo_param_to_solver_param_map = dict()
        self._needs_updated = True
        self._last_results_object: Optional[WntrResults] = None
        self._pyomo_to_wntr_visitor = PyomoToWntrVisitor(
            self._pyomo_var_to_solver_var_map, self._pyomo_param_to_solver_param_map
        )

    def available(self):
        if wntr_available:
            return self.Availability.FullLicense
        else:
            return self.Availability.NotFound

    def version(self):
        return tuple(int(i) for i in wntr.__version__.split('.'))

    @property
    def config(self) -> WntrConfig:
        return self._config

    @config.setter
    def config(self, val: WntrConfig):
        self._config = val

    @property
    def wntr_options(self):
        return self._solver_options

    @wntr_options.setter
    def wntr_options(self, val: Dict):
        self._solver_options = val

    @property
    def symbol_map(self):
        return self._symbol_map

    def _solve(self, timer: HierarchicalTimer):
        options = dict()
        if self.config.time_limit is not None:
            options['TIME_LIMIT'] = self.config.time_limit
        options.update(self.wntr_options)
        opt = wntr.sim.solvers.NewtonSolver(options)

        if self.config.stream_solver:
            ostream = sys.stdout
        else:
            ostream = None

        t0 = time.time()
        if self._needs_updated:
            timer.start('set_structure')
            self._solver_model.set_structure()
            timer.stop('set_structure')
            self._needs_updated = False
        timer.start('newton solve')
        status, msg, num_iter = opt.solve(self._solver_model, ostream)
        timer.stop('newton solve')
        tf = time.time()

        results = WntrResults(self)
        results.wallclock_time = tf - t0
        if status == wntr.sim.solvers.SolverStatus.converged:
            results.termination_condition = TerminationCondition.optimal
        else:
            results.termination_condition = TerminationCondition.error
        results.best_feasible_objective = None
        results.best_objective_bound = None

        if self.config.load_solution:
            if status == wntr.sim.solvers.SolverStatus.converged:
                timer.start('load solution')
                self.load_vars()
                timer.stop('load solution')
            else:
                raise RuntimeError(
                    'A feasible solution was not found, so no solution can be loaded. '
                    'If using the appsi.solvers.Wntr interface, you can '
                    'set opt.config.load_solution=False. If using the environ.SolverFactory '
                    'interface, you can set opt.solve(model, load_solutions = False). '
                    'Then you can check results.termination_condition and '
                    'results.best_feasible_objective before loading a solution.'
                )
        return results

    def solve(self, model: BlockData, timer: HierarchicalTimer = None) -> Results:
        StaleFlagManager.mark_all_as_stale()
        if self._last_results_object is not None:
            self._last_results_object.solution_loader.invalidate()
        if timer is None:
            timer = HierarchicalTimer()
        if model is not self._model:
            timer.start('set_instance')
            self.set_instance(model)
            timer.stop('set_instance')
        else:
            timer.start('update')
            self.update(timer=timer)
            timer.start('initial values')
            for v_id, solver_v in self._pyomo_var_to_solver_var_map.items():
                pyomo_v = self._vars[v_id][0]
                val = pyomo_v.value
                if val is not None:
                    solver_v.value = val
            timer.stop('initial values')
            timer.stop('update')
        res = self._solve(timer)
        self._last_results_object = res
        if self.config.report_timing:
            logger.info('\n' + str(timer))
        return res

    def _reinit(self):
        saved_config = self.config
        saved_options = self.wntr_options
        saved_update_config = self.update_config
        self.__init__(only_child_vars=self._only_child_vars)
        self.config = saved_config
        self.wntr_options = saved_options
        self.update_config = saved_update_config

    def set_instance(self, model):
        if self._last_results_object is not None:
            self._last_results_object.solution_loader.invalidate()
        if not self.available():
            c = self.__class__
            raise PyomoException(
                f'Solver {c.__module__}.{c.__qualname__} is not available '
                f'({self.available()}).'
            )
        self._reinit()
        self._model = model
        if self.use_extensions and cmodel_available:
            self._expr_types = cmodel.PyomoExprTypes()

        if self.config.symbolic_solver_labels:
            self._labeler = TextLabeler()
        else:
            self._labeler = NumericLabeler('x')

        self._solver_model = wntr.sim.aml.aml.Model()
        self._solver_model._wntr_fixed_var_params = wntr.sim.aml.aml.ParamDict()
        self._solver_model._wntr_fixed_var_cons = wntr.sim.aml.aml.ConstraintDict()

        self.add_block(model)

    def _add_variables(self, variables: List[VarData]):
        aml = wntr.sim.aml.aml
        for var in variables:
            varname = self._symbol_map.getSymbol(var, self._labeler)
            _v, _lb, _ub, _fixed, _domain_interval, _value = self._vars[id(var)]
            lb, ub, step = _domain_interval
            if (
                _lb is not None
                or _ub is not None
                or lb is not None
                or ub is not None
                or step != 0
            ):
                raise ValueError(
                    f"WNTR's newton solver only supports continuous variables without bounds: {var.name}"
                )
            if _value is None:
                _value = 0
            wntr_var = aml.Var(_value)
            setattr(self._solver_model, varname, wntr_var)
            self._pyomo_var_to_solver_var_map[id(var)] = wntr_var
            if _fixed:
                self._solver_model._wntr_fixed_var_params[id(var)] = param = aml.Param(
                    _value
                )
                wntr_expr = wntr_var - param
                self._solver_model._wntr_fixed_var_cons[id(var)] = aml.Constraint(
                    wntr_expr
                )
            self._needs_updated = True

    def _add_params(self, params: List[ParamData]):
        aml = wntr.sim.aml.aml
        for p in params:
            pname = self._symbol_map.getSymbol(p, self._labeler)
            wntr_p = aml.Param(p.value)
            setattr(self._solver_model, pname, wntr_p)
            self._pyomo_param_to_solver_param_map[id(p)] = wntr_p

    def _add_constraints(self, cons: List[ConstraintData]):
        aml = wntr.sim.aml.aml
        for con in cons:
            if not con.equality:
                raise ValueError(
                    f"WNTR's newtwon solver only supports equality constraints: {con.name}"
                )
            conname = self._symbol_map.getSymbol(con, self._labeler)
            wntr_expr = self._pyomo_to_wntr_visitor.dfs_postorder_stack(
                con.body - con.upper
            )
            wntr_con = aml.Constraint(wntr_expr)
            setattr(self._solver_model, conname, wntr_con)
            self._pyomo_con_to_solver_con_map[con] = wntr_con
            self._needs_updated = True

    def _remove_constraints(self, cons: List[ConstraintData]):
        for con in cons:
            solver_con = self._pyomo_con_to_solver_con_map[con]
            delattr(self._solver_model, solver_con.name)
            self._symbol_map.removeSymbol(con)
            del self._pyomo_con_to_solver_con_map[con]
            self._needs_updated = True

    def _remove_variables(self, variables: List[VarData]):
        for var in variables:
            v_id = id(var)
            solver_var = self._pyomo_var_to_solver_var_map[v_id]
            delattr(self._solver_model, solver_var.name)
            self._symbol_map.removeSymbol(var)
            del self._pyomo_var_to_solver_var_map[v_id]
            if v_id in self._solver_model._wntr_fixed_var_params:
                del self._solver_model._wntr_fixed_var_params[v_id]
                del self._solver_model._wntr_fixed_var_cons[v_id]
            self._needs_updated = True

    def _remove_params(self, params: List[ParamData]):
        for p in params:
            p_id = id(p)
            solver_param = self._pyomo_param_to_solver_param_map[p_id]
            delattr(self._solver_model, solver_param.name)
            self._symbol_map.removeSymbol(p)
            del self._pyomo_param_to_solver_param_map[p_id]

    def _update_variables(self, variables: List[VarData]):
        aml = wntr.sim.aml.aml
        for var in variables:
            v_id = id(var)
            solver_var = self._pyomo_var_to_solver_var_map[v_id]
            _v, _lb, _ub, _fixed, _domain_interval, _value = self._vars[v_id]
            lb, ub, step = _domain_interval
            if (
                _lb is not None
                or _ub is not None
                or lb is not None
                or ub is not None
                or step != 0
            ):
                raise ValueError(
                    f"WNTR's newton solver only supports continuous variables without bounds: {var.name}"
                )
            if _value is None:
                _value = 0
            solver_var.value = _value
            if _fixed:
                if v_id not in self._solver_model._wntr_fixed_var_params:
                    self._solver_model._wntr_fixed_var_params[v_id] = param = aml.Param(
                        _value
                    )
                    wntr_expr = solver_var - param
                    self._solver_model._wntr_fixed_var_cons[v_id] = aml.Constraint(
                        wntr_expr
                    )
                    self._needs_updated = True
                else:
                    self._solver_model._wntr_fixed_var_params[v_id].value = _value
            else:
                if v_id in self._solver_model._wntr_fixed_var_params:
                    del self._solver_model._wntr_fixed_var_params[v_id]
                    del self._solver_model._wntr_fixed_var_cons[v_id]
                    self._needs_updated = True

    def update_params(self):
        for p_id, solver_p in self._pyomo_param_to_solver_param_map.items():
            p = self._params[p_id]
            solver_p.value = p.value

    def _set_objective(self, obj):
        raise NotImplementedError(
            f"WNTR's newton solver can only solve square problems"
        )

    def load_vars(self, vars_to_load=None):
        if vars_to_load is None:
            vars_to_load = [i[0] for i in self._vars.values()]
        for v in vars_to_load:
            v_id = id(v)
            solver_v = self._pyomo_var_to_solver_var_map[v_id]
            v.value = solver_v.value

    def get_primals(self, vars_to_load=None):
        if vars_to_load is None:
            vars_to_load = [i[0] for i in self._vars.values()]
        res = ComponentMap()
        for v in vars_to_load:
            v_id = id(v)
            solver_v = self._pyomo_var_to_solver_var_map[v_id]
            res[v] = solver_v.value
        return res

    def _add_sos_constraints(self, cons):
        if len(cons) > 0:
            raise NotImplementedError(
                f"WNTR's newton solver does not support SOS constraints"
            )

    def _remove_sos_constraints(self, cons):
        if len(cons) > 0:
            raise NotImplementedError(
                f"WNTR's newton solver does not support SOS constraints"
            )


def _handle_product_expression(node, values):
    arg1, arg2 = values
    return arg1 * arg2


def _handle_sum_expression(node, values):
    return sum(values)


def _handle_division_expression(node, values):
    arg1, arg2 = values
    return arg1 / arg2


def _handle_pow_expression(node, values):
    arg1, arg2 = values
    return arg1**arg2


def _handle_negation_expression(node, values):
    return -values[0]


def _handle_exp_expression(node, values):
    return wntr.sim.aml.exp(values[0])


def _handle_log_expression(node, values):
    return wntr.sim.aml.log(values[0])


def _handle_sin_expression(node, values):
    return wntr.sim.aml.sin(values[0])


def _handle_cos_expression(node, values):
    return wntr.sim.aml.cos(values[0])


def _handle_tan_expression(node, values):
    return wntr.sim.aml.tan(values[0])


def _handle_asin_expression(node, values):
    return wntr.sim.aml.asin(values[0])


def _handle_acos_expression(node, values):
    return wntr.sim.aml.acos(values[0])


def _handle_atan_expression(node, values):
    return wntr.sim.aml.atan(values[0])


def _handle_sqrt_expression(node, values):
    return (values[0]) ** 0.5


def _handle_abs_expression(node, values):
    return wntr.sim.aml.abs(values[0])


_unary_handler_map = dict()
_unary_handler_map['exp'] = _handle_exp_expression
_unary_handler_map['log'] = _handle_log_expression
_unary_handler_map['sin'] = _handle_sin_expression
_unary_handler_map['cos'] = _handle_cos_expression
_unary_handler_map['tan'] = _handle_tan_expression
_unary_handler_map['asin'] = _handle_asin_expression
_unary_handler_map['acos'] = _handle_acos_expression
_unary_handler_map['atan'] = _handle_atan_expression
_unary_handler_map['sqrt'] = _handle_sqrt_expression
_unary_handler_map['abs'] = _handle_abs_expression


def _handle_unary_function_expression(node, values):
    if node.getname() in _unary_handler_map:
        return _unary_handler_map[node.getname()](node, values)
    else:
        raise NotImplementedError(
            f'Unrecognized unary function expression: {node.getname()}'
        )


_handler_map = dict()
_handler_map[ProductExpression] = _handle_product_expression
_handler_map[DivisionExpression] = _handle_division_expression
_handler_map[PowExpression] = _handle_pow_expression
_handler_map[SumExpression] = _handle_sum_expression
_handler_map[MonomialTermExpression] = _handle_product_expression
_handler_map[NegationExpression] = _handle_negation_expression
_handler_map[UnaryFunctionExpression] = _handle_unary_function_expression
_handler_map[LinearExpression] = _handle_sum_expression
_handler_map[AbsExpression] = _handle_abs_expression
_handler_map[NPV_ProductExpression] = _handle_product_expression
_handler_map[NPV_DivisionExpression] = _handle_division_expression
_handler_map[NPV_PowExpression] = _handle_pow_expression
_handler_map[NPV_SumExpression] = _handle_sum_expression
_handler_map[NPV_NegationExpression] = _handle_negation_expression
_handler_map[NPV_UnaryFunctionExpression] = _handle_unary_function_expression
_handler_map[NPV_AbsExpression] = _handle_abs_expression


class PyomoToWntrVisitor(ExpressionValueVisitor):
    def __init__(self, var_map, param_map):
        self.var_map = var_map
        self.param_map = param_map

    def visit(self, node, values):
        if node.__class__ in _handler_map:
            return _handler_map[node.__class__](node, values)
        else:
            raise NotImplementedError(f'Unrecognized expression type: {node.__class__}')

    def visiting_potential_leaf(self, node):
        if node.__class__ in native_numeric_types:
            return True, node

        if node.is_variable_type():
            return True, self.var_map[id(node)]

        if node.is_parameter_type():
            return True, self.param_map[id(node)]

        return False, None



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import types
import weakref
from pyomo.common.pyomo_typing import overload

from ctypes import (
    Structure,
    POINTER,
    CFUNCTYPE,
    cdll,
    byref,
    c_int,
    c_long,
    c_ulong,
    c_double,
    c_byte,
    c_char_p,
    c_void_p,
)

from pyomo.common.autoslots import AutoSlots
from pyomo.common.fileutils import find_library
from pyomo.common.numeric_types import (
    check_if_native_type,
    native_types,
    native_numeric_types,
    value,
    _pyomo_constant_types,
)
from pyomo.core.expr.numvalue import NonNumericValue, NumericConstant
import pyomo.core.expr as EXPR
from pyomo.core.base.component import Component
from pyomo.core.base.units_container import units

logger = logging.getLogger('pyomo.core')
nan = float('nan')


class ExternalFunction(Component):
    """Interface to an external (non-algebraic) function.

    :class:`ExternalFunction` provides an interface for declaring
    general user-provided functions, and then embedding calls to the
    external functions within Pyomo expressions.

    .. note::

       Just because you can express a Pyomo model with external
       functions does not mean that the resulting model is solvable.  In
       particular, linear solvers do not accept external functions.  The
       AMPL Solver Library (ASL) interface does support external
       functions for general nonlinear solvers compiled against it, but
       only allows functions in compiled libraries through the
       :class:`AMPLExternalFunction` interface.

    """

    def __new__(cls, *args, **kwargs):
        if cls is not ExternalFunction:
            return super().__new__(cls)
        elif args:
            return super().__new__(PythonCallbackFunction)
        elif 'library' not in kwargs and any(
            kw in kwargs for kw in ('function', 'fgh')
        ):
            return super().__new__(PythonCallbackFunction)
        else:
            return super().__new__(AMPLExternalFunction)

    @overload
    def __init__(self, function=None, gradient=None, hessian=None, *, fgh=None): ...

    @overload
    def __init__(self, *, library: str, function: str): ...

    def __init__(self, *args, **kwargs):
        """Construct a reference to an external function.

        There are two fundamental interfaces supported by
        :class:`ExternalFunction`: Python callback functions and AMPL
        external functions.

        **Python callback functions** (:class:`PythonCallbackFunction`
        interface)

        Python callback functions can be specified one of two ways:

        1. FGH interface:

          A single external function call with a signature matching the
          :meth:`evaluate_fgh()` method.

        2. Independent functions:

          One to three functions that can evaluate the function value,
          gradient of the function [partial derivatives] with respect to
          its inputs, and the Hessian of the function [partial second
          derivatives].  The ``function`` interface expects a function
          matching the prototype:

          .. code::

             def function(*args): float

          The ``gradient`` and ``hessian`` interface expect functions
          matching the prototype:

          .. code::

             def gradient_or_hessian(args, fixed=None): List[float]

          Where ``args`` is a tuple of function arguments and ``fixed``
          is either None or a list of values equal in length to ``args``
          indicating which arguments are currently fixed (``True``) or
          variable (``False``).

        **ASL function libraries** (:class:`AMPLExternalFunction` interface)

        Pyomo can also call functions compiled as part of an AMPL
        External Function library (see the `Imported functions` section
        in the `Hooking your solver to AMPL
        <https://www.ampl.com/REFS/hooking3.pdf>`_ report).  Links to
        these functions are declared by creating an
        :class:`ExternalFunction` and passing the compiled library name
        (or path) to the ``library`` keyword and the name of the
        function to the ``function`` keyword.

        """
        self._units = kwargs.pop('units', None)
        if self._units is not None:
            self._units = units.get_units(self._units)
        self._arg_units = kwargs.pop('arg_units', None)
        if self._arg_units is not None:
            self._arg_units = [units.get_units(u) for u in self._arg_units]
        kwargs.setdefault('ctype', ExternalFunction)
        Component.__init__(self, **kwargs)
        self._constructed = True
        ### HACK ###
        # FIXME: We must declare an _index attribute because
        # block._add_temporary_set assumes ALL components define an
        # index.  Sigh.
        self._index_set = None

    def get_units(self):
        """Return the units for this ExternalFunction"""
        return self._units

    def get_arg_units(self):
        """Return the units for this ExternalFunctions arguments"""
        return self._arg_units

    def __call__(self, *args):
        """Return a Pyomo expression that evaluates this function

        Note that this does not evaluate the external function
        represented by this :class:`ExternalFunction` component, but
        rather generates a Pyomo expression object that represents
        symbolic evaluation of the external function at the specified
        arguments (which themselves may be Pyomo expression objects).

        Parameters
        ----------
        *args:
            Arguments (constants, variables, expressions) to pass to
            this External function.

        Returns
        -------
        ExternalFunctionExpression or NPV_ExternalFunctionExpression:
            The new Pyomo expression node.

        """

        args_ = []
        for arg in args:
            if type(arg) is types.GeneratorType:
                args_.extend(arg)
            else:
                args_.append(arg)
        #
        # Loop and do two thing:
        #   1. Wrap non-numeric arguments
        #   2. See if we have a potentially variable argument
        #
        pv = False
        for i, arg in enumerate(args_):
            try:
                if arg.__class__ in native_types:
                    continue
                if arg.is_potentially_variable():
                    pv = True
                continue
            except AttributeError:
                if check_if_native_type(arg):
                    continue
            args_[i] = NonNumericValue(arg)
        #
        if pv:
            return EXPR.ExternalFunctionExpression(args_, self)
        return EXPR.NPV_ExternalFunctionExpression(args_, self)

    def evaluate(self, args):
        """Return the value of the function given the specified arguments

        Parameters
        ----------
        args: Iterable
            Iterable containing the arguments to pass to the external
            function.  Non-native type elements will be converted to a
            native value using the :py:func:`value()` function.

        Returns
        -------
        float
            The return value of the function evaluated at `args`
        """
        args_ = [arg if arg.__class__ in native_types else value(arg) for arg in args]
        return self._evaluate(args_, None, 0)[0]

    def evaluate_fgh(self, args, fixed=None, fgh=2):
        """Evaluate the function and gradients given the specified arguments

        This evaluates the function given the specified arguments
        returning a 3-tuple of (function value [f], list of first partial
        derivatives [g], and the upper triangle of the Hessian matrix
        [h]).

        Parameters
        ----------
        args: Iterable
            Iterable containing the arguments to pass to the external
            function.  Non-native type elements will be converted to a
            native value using the :py:func:`value()` function.

        fixed: Optional[List[bool]]
            List of values indicating if the corresponding argument
            value is fixed.  Any fixed indices are guaranteed to return
            0 for first and second derivatives, regardless of what is
            computed by the external function.

        fgh: {0, 1, 2}
            What evaluations to return:

            * **0**: just return function evaluation
            * **1**: return function and first derivatives
            * **2**: return function, first derivatives, and hessian matrix

            Any return values not requested will be `None`.

        Returns
        -------
        f: float
            The return value of the function evaluated at `args`
        g: List[float] or None
            The list of first partial derivatives
        h: List[float] or None
            The upper-triangle of the Hessian matrix (second partial
            derivatives), stored column-wise.  Element :math:`H_{i,j}`
            (with :math:`0 <= i <= j < N` are mapped using
            :math:`h[i + j*(j + 1)/2] == H_{i,j}`.

        """
        args_ = [arg if arg.__class__ in native_types else value(arg) for arg in args]
        # Note: this is passed-by-reference, and the args_ list may be
        # changed by _evaluate (e.g., for PythonCallbackFunction).
        # Remember the original length of the list.
        N = len(args_)
        f, g, h = self._evaluate(args_, fixed, fgh)
        # Guarantee the return value behavior documented in the docstring
        if fgh == 2:
            n = N - 1
            if len(h) - 1 != n + n * (n + 1) // 2:
                raise RuntimeError(
                    f"External function '{self.name}' returned an invalid "
                    f"Hessian matrix (expected {n + n*(n+1)//2 + 1}, "
                    f"received {len(h)})"
                )
        else:
            h = None
        if fgh >= 1:
            if len(g) != N:
                raise RuntimeError(
                    f"External function '{self.name}' returned an invalid "
                    f"derivative vector (expected {N}, "
                    f"received {len(g)})"
                )
        else:
            g = None
        # Note: the ASL does not require clients to honor the fixed flag
        # (allowing them to return non-0 values for the derivative with
        # respect to a fixed numeric value).  We will allow clients to
        # be similarly lazy and enforce the fixed flag here.
        if fixed is not None:
            if fgh >= 1:
                for i, v in enumerate(fixed):
                    if not v:
                        continue
                    g[i] = 0
            if fgh >= 2:
                for i, v in enumerate(fixed):
                    if not v:
                        continue
                    for j in range(N):
                        if i <= j:
                            h[i + (j * (j + 1)) // 2] = 0
                        else:
                            h[j + (i * (i + 1)) // 2] = 0
        return f, g, h

    def _evaluate(self, args, fixed, fgh):
        """Derived class implementation to evaluate an external function.

        The API follows :meth:`evaluate_fgh()`, with the exception that
        the `args` iterable is guaranteed to already have been reduced
        to concrete values (it will not have Pyomo components or
        expressions).
        """
        raise NotImplementedError(f"{type(self)} did not implement _evaluate()")


class AMPLExternalFunction(ExternalFunction):
    __autoslot_mappers__ = {
        # Remove reference to loaded library (they are not copyable or
        # picklable)
        '_so': AutoSlots.encode_as_none,
        '_known_functions': AutoSlots.encode_as_none,
    }

    def __init__(self, *args, **kwargs):
        if args:
            raise ValueError(
                "AMPLExternalFunction constructor does not support "
                "positional arguments"
            )
        self._library = kwargs.pop('library', None)
        self._function = kwargs.pop('function', None)
        self._known_functions = None
        self._so = None
        # Convert the specified library name to an absolute path, and
        # warn the user if we couldn't find it
        if self._library is not None:
            _lib = find_library(self._library)
            if _lib is not None:
                self._library = _lib
            else:
                logger.warning(
                    'Defining AMPL external function, but cannot locate '
                    f'specified library "{self._library}"'
                )
        ExternalFunction.__init__(self, *args, **kwargs)

    def _evaluate(self, args, fixed, fgh):
        if self._so is None:
            self.load_library()
        if self._function not in self._known_functions:
            raise RuntimeError(
                "Error: external function '%s' was not registered within "
                "external library %s.\n\tAvailable functions: (%s)"
                % (
                    self._function,
                    self._library,
                    ', '.join(self._known_functions.keys()),
                )
            )
        #
        N = len(args)
        arglist = _ARGLIST(args, fgh, fixed)
        fcn = self._known_functions[self._function][0]
        f = fcn(byref(arglist))
        if fgh >= 1:
            g = [nan] * N
            for i in range(N):
                if arglist.at[i] < 0:
                    continue
                g[i] = arglist.derivs[arglist.at[i]]
        else:
            g = None
        if fgh >= 2:
            h = [nan] * ((N + N**2) // 2)
            for j in range(N):
                j_r = arglist.at[j]
                if j_r < 0:
                    continue
                for i in range(j + 1):
                    i_r = arglist.at[i]
                    if i_r < 0:
                        continue
                    h[i + j * (j + 1) // 2] = arglist.hes[i_r + j_r * (j_r + 1) // 2]
        else:
            h = None
        return f, g, h

    def load_library(self):
        # Note that the library was located in __init__ and converted to
        # an absolute path (including checking the CWD).  However, we
        # previously tested that changing the environment (i.e.,
        # changing directories) between defining the ExternalFunction
        # and loading it would cause the library to still be correctly
        # loaded.  We will re-search for the library here.  If it was
        # previously found, we will be searching for an absolute path
        # (and the same path will be found again).
        _abs_lib = find_library(self._library)
        if _abs_lib is not None:
            self._library = _abs_lib
        self._so = cdll.LoadLibrary(self._library)
        self._known_functions = {}
        AE = _AMPLEXPORTS()
        AE.ASLdate = 20160307

        def addfunc(name, f, _type, nargs, funcinfo, ae):
            # trap for Python 3, where the name comes in as bytes() and
            # not a string
            if not isinstance(name, str):
                name = name.decode()
            self._known_functions[str(name)] = (f, _type, nargs, funcinfo, ae)

        AE.Addfunc = _AMPLEXPORTS.ADDFUNC(addfunc)

        def addrandinit(ae, rss, v):
            # TODO: This should support the randinit ASL option
            rss(v, 1)

        AE.Addrandinit = _AMPLEXPORTS.ADDRANDINIT(addrandinit)

        def atreset(ae, a, b):
            logger.warning(
                "AMPL External function: ignoring AtReset call in external "
                "library.  This may result in a memory leak or other "
                "undesirable behavior."
            )

        AE.AtReset = _AMPLEXPORTS.ATRESET(atreset)

        FUNCADD = CFUNCTYPE(None, POINTER(_AMPLEXPORTS))
        FUNCADD(('funcadd_ASL', self._so))(byref(AE))

    def _pprint(self):
        return (
            [
                ('function', self._function),
                ('library', self._library),
                ('units', str(self._units)),
                (
                    'arg_units',
                    (
                        [str(u) for u in self._arg_units]
                        if self._arg_units is not None
                        else None
                    ),
                ),
            ],
            None,
            None,
            None,
        )


def _python_callback_fid_mapper(encode, val):
    if encode:
        return PythonCallbackFunction.global_registry[val]()
    else:
        return PythonCallbackFunction.register_instance(val)


class _PythonCallbackFunctionID(NumericConstant):
    """A specialized NumericConstant to preserve FunctionIDs through deepcopy.

    Function IDs are effectively pointers back to the
    PythonCallbackFunction.  As such, we need special handling to
    maintain / preserve the correct linkages through deepcopy (and
    model.clone()).

    """

    __slots__ = ()
    __autoslot_mappers__ = {'value': _python_callback_fid_mapper}

    def is_constant(self):
        # Return False so this object is not simplified out of expressions
        return False


_pyomo_constant_types.add(_PythonCallbackFunctionID)


class PythonCallbackFunction(ExternalFunction):
    __autoslot_mappers__ = {'_fcn_id': _python_callback_fid_mapper}

    global_registry = []
    global_id_to_fid = {}

    @classmethod
    def register_instance(cls, instance):
        # Ideally, we would just use the id() as the function id, but
        # Python id() is more than 32-bits, which can cause problems for
        # the ASL.  We will map id() to a "small" integer and use that
        # small integer as the "function id".
        _id = id(instance)
        if _id in cls.global_id_to_fid:
            prev_fid = cls.global_id_to_fid[_id]
            prev_obj = cls.global_registry[prev_fid]
            if prev_obj() is not None:
                assert prev_obj() is instance
                return prev_fid
        cls.global_id_to_fid[_id] = _fid = len(cls.global_registry)
        cls.global_registry.append(weakref.ref(instance))
        return _fid

    def __init__(self, *args, **kwargs):
        for i, kw in enumerate(('function', 'gradient', 'hessian')):
            if len(args) <= i:
                break
            if kw in kwargs:
                raise ValueError(
                    "Duplicate definition of external function through "
                    f"positional and keyword ('{kw}=') arguments"
                )
            kwargs[kw] = args[i]
        if len(args) > 3:
            raise ValueError(
                "PythonCallbackFunction constructor only supports "
                "0 - 3 positional arguments"
            )
        self._fcn = kwargs.pop('function', None)
        self._grad = kwargs.pop('gradient', None)
        self._hess = kwargs.pop('hessian', None)
        self._fgh = kwargs.pop('fgh', None)
        if self._fgh is not None and any((self._fcn, self._grad, self._hess)):
            raise ValueError(
                "Cannot specify 'fgh' with any of {'function', 'gradient', hessian'}"
            )

        # There is an implicit first argument (the function pointer), we
        # need to add that to the arg_units
        arg_units = kwargs.get('arg_units', None)
        if arg_units is not None:
            kwargs['arg_units'] = list(arg_units)
            kwargs['arg_units'].append(None)

        # TODO: complete/distribute the pyomo_socket_server /
        # pyomo_ampl.so AMPL extension
        self._library = 'pyomo_ampl.so'
        self._function = 'pyomo_socket_server'
        ExternalFunction.__init__(self, *args, **kwargs)
        self._fcn_id = PythonCallbackFunction.register_instance(self)

    def __call__(self, *args):
        # NOTE: we append the Function ID to the END of the argument
        # list because it is easier to update the gradient / hessian
        # results
        return super().__call__(*args, _PythonCallbackFunctionID(self._fcn_id))

    def _evaluate(self, args, fixed, fgh):
        # Remove the fcn_id
        _id = args.pop()
        if fixed is not None:
            fixed = fixed[:-1]
        if _id != self._fcn_id:
            raise RuntimeError("PythonCallbackFunction called with invalid Global ID")
        if self._fgh is not None:
            f, g, h = self._fgh(args, fgh, fixed)
        else:
            f = self._fcn(*args)
            if fgh >= 1:
                if self._grad is None:
                    raise RuntimeError(
                        f"ExternalFunction '{self.name}' was not defined "
                        "with a gradient callback.  Cannot evaluate the "
                        "derivative of the function"
                    )
                g = self._grad(args, fixed)
            else:
                g = None
            if fgh == 2:
                if self._hess is None:
                    raise RuntimeError(
                        f"ExternalFunction '{self.name}' was not defined "
                        "with a Hessian callback.  Cannot evaluate the "
                        "second derivative of the function"
                    )
                h = self._hess(args, fixed)
            else:
                h = None
        # Update g,h to reflect the function id (which by definition is
        # always fixed)
        if g is not None:
            g.append(0)
        if h is not None:
            h.extend([0] * (len(args) + 1))
        return f, g, h

    def _pprint(self):
        return (
            [
                ('function', self._fcn.__qualname__),
                ('units', str(self._units)),
                (
                    'arg_units',
                    (
                        [str(u) for u in self._arg_units[:-1]]
                        if self._arg_units is not None
                        else None
                    ),
                ),
            ],
            None,
            None,
            None,
        )


class _ARGLIST(Structure):
    """Mock up the arglist structure from AMPL's funcadd.h

    This data structure is populated by AMPL when calling external
    functions (for both passing in the function arguments and retrieving
    the derivative/Hessian information.
    """

    _fields_ = [
        ('n', c_int),  # number of args
        ('nr', c_int),  # number of real input args
        ('at', POINTER(c_int)),  # argument types -- see DISCUSSION below
        ('ra', POINTER(c_double)),  # pure real args (IN, OUT, and INOUT)
        ('sa', POINTER(c_char_p)),  # symbolic IN args
        ('derivs', POINTER(c_double)),  # for partial derivatives (if nonzero)
        ('hes', POINTER(c_double)),  # for second partials (if nonzero)
        # if (dig && dig[i]) { partials w.r.t. ra[i] will not be used }
        ('dig', POINTER(c_byte)),
        ('funcinfo', c_char_p),  # for use by the function (if desired)
        ('AE', c_void_p),  # functions made visible (via #defines below)
        ('f', c_void_p),  # for internal use by AMPL
        ('tva', c_void_p),  # for internal use by AMPL
        ('Errmsg', c_char_p),  # To indicate an error, set this to a
        # description of the error.  When derivs is nonzero and the
        # error is that first derivatives cannot or are not computed, a
        # single quote character (') should be the first character in
        # the text assigned to Errmsg, followed by the actual error
        # message.  Similarly, if hes is nonzero and the error is that
        # second derivatives are not or cannot be computed, a double
        # quote character (") should be the first character in Errmsg,
        # followed by the actual error message text.
        ('TMI', c_void_p),  # used in Tempmem calls
        ('Private', c_char_p),
        # The following fields are relevant only when imported functions
        # are called by AMPL commands (not declarations).
        ('nin', c_int),  # number of input (IN and INOUT) args
        ('nout', c_int),  # number of output (OUT and INOUT) args
        ('nsin', c_int),  # number of symbolic input arguments
        ('nsout', c_int),  # number of symbolic OUT and INOUT args
    ]

    def __init__(self, args, fgh=0, fixed=None):
        super().__init__()
        self._encoded_strings = []
        self.n = len(args)
        self.at = (c_int * self.n)()
        _reals = []
        _strings = []
        nr = 0
        ns = 0
        for i, arg in enumerate(args):
            if arg.__class__ in native_numeric_types:
                _reals.append(arg)
                self.at[i] = nr
                nr += 1
                continue
            if isinstance(arg, str):
                # String arguments need to be converted to bytes
                # (encoded as plain ASCII characters).  We will
                # explicitly cache the resulting bytes object to make
                # absolutely sure that the Python garbage collector
                # doesn't accidentally clean up the object before we
                # call the external function.
                arg = arg.encode('ascii')
                self._encoded_strings.append(arg)
            if isinstance(arg, bytes):
                _strings.append(arg)
                # Note increment first, as at[i] starts at -1 for strings
                ns += 1
                self.at[i] = -ns
            else:
                raise RuntimeError(
                    f"Unknown data type, {type(arg).__name__}, passed as "
                    f"argument {i} for an ASL ExternalFunction"
                )
        self.nr = nr
        self.ra = (c_double * nr)(*_reals)
        self.sa = (c_char_p * ns)(*_strings)
        if fgh >= 1:
            self.derivs = (c_double * nr)(0.0)
        if fgh >= 2:
            self.hes = (c_double * ((nr + nr * nr) // 2))(0.0)

        if fixed:
            self.dig = (c_byte * nr)(0)
            for i, v in enumerate(fixed):
                if v:
                    r_idx = self.at[i]
                    if r_idx >= 0:
                        self.dig[r_idx] = 1


# The following "fake" class resolves a circular reference issue in the
# _AMPLEXPORTS datastructure
#
# Declare a dummy structure called _AMPLEXPORTS so that the symbol exists
# and the POINTER(_AMPLEXPORTS) call in the definition of ADDFUNC (needed
# to define AMPLExPORTS) will succeed.
class _AMPLEXPORTS(Structure):
    pass


class _AMPLEXPORTS(Structure):
    """Mock up the AmplExports structure from AMPL's funcadd.h

    The only thing we really need here is to be able to override the
    Addfunc function pointer to point back into Python so we can
    intercept the registration of external AMPL functions.  Ideally, we
    would populate the other function pointers as well, but that is
    trickier than it sounds, and at least so far is completely unneeded.
    """

    AMPLFUNC = CFUNCTYPE(c_double, POINTER(_ARGLIST))

    ADDFUNC = CFUNCTYPE(
        None, c_char_p, AMPLFUNC, c_int, c_int, c_void_p, POINTER(_AMPLEXPORTS)
    )

    RANDSEEDSETTER = CFUNCTYPE(None, c_void_p, c_ulong)

    ADDRANDINIT = CFUNCTYPE(None, POINTER(_AMPLEXPORTS), RANDSEEDSETTER, c_void_p)

    ATRESET = CFUNCTYPE(None, POINTER(_AMPLEXPORTS), c_void_p, c_void_p)

    _fields_ = [
        ('StdErr', c_void_p),
        ('Addfunc', ADDFUNC),
        ('ASLdate', c_long),
        ('FprintF', c_void_p),
        ('PrintF', c_void_p),
        ('SprintF', c_void_p),
        ('VfprintF', c_void_p),
        ('VsprintF', c_void_p),
        ('Strtod', c_void_p),
        ('Crypto', c_void_p),
        ('asl', c_char_p),
        ('AtExit', c_void_p),
        ('AtReset', ATRESET),
        ('Tempmem', c_void_p),
        ('Add_table_handler', c_void_p),
        ('Private', c_char_p),
        ('Qsortv', c_void_p),
        # /* More stuff for stdio in DLLs... */
        ('StdIn', c_void_p),
        ('StdOut', c_void_p),
        ('Clearerr', c_void_p),
        ('Fclose', c_void_p),
        ('Fdopen', c_void_p),
        ('Feof', c_void_p),
        ('Ferror', c_void_p),
        ('Fflush', c_void_p),
        ('Fgetc', c_void_p),
        ('Fgets', c_void_p),
        ('Fileno', c_void_p),
        ('Fopen', c_void_p),
        ('Fputc', c_void_p),
        ('Fputs', c_void_p),
        ('Fread', c_void_p),
        ('Freopen', c_void_p),
        ('Fscanf', c_void_p),
        ('Fseek', c_void_p),
        ('Ftell', c_void_p),
        ('Fwrite', c_void_p),
        ('Pclose', c_void_p),
        ('Perror', c_void_p),
        ('Popen', c_void_p),
        ('Puts', c_void_p),
        ('Rewind', c_void_p),
        ('Scanf', c_void_p),
        ('Setbuf', c_void_p),
        ('Setvbuf', c_void_p),
        ('Sscanf', c_void_p),
        ('Tempnam', c_void_p),
        ('Tmpfile', c_void_p),
        ('Tmpnam', c_void_p),
        ('Ungetc', c_void_p),
        ('AI', c_void_p),
        ('Getenv', c_void_p),
        ('Breakfunc', c_void_p),
        ('Breakarg', c_char_p),
        # /* Items available with ASLdate >= 20020501 start here. */
        ('SnprintF', c_void_p),
        ('VsnprintF', c_void_p),
        ('Addrand', c_void_p),
        ('Addrandinit', ADDRANDINIT),
    ]
